# roms/_nonparametric.py
"""Nonparametric ROM class."""

__all__ = [
    "ROM",
]

import numpy as np

from ..models import _utils as modutils
from ._base import _BaseROM


class ROM(_BaseROM):
    r"""Nonparametric reduced-order model.

    This class connects classes from the various submodules to form a complete
    reduced-order modeling workflow.

    High-dimensional data
    :math:`\to` transformed / preprocessed data
    :math:`\to` compressed data
    :math:`\to` low-dimensional model.

    Parameters
    ----------
    model : :mod:`opinf.models` object
        Nonparametric system model, an instance of one of the following:

        * :class:`opinf.models.ContinuousModel`
        * :class:`opinf.models.DiscreteModel`
    lifter : :mod:`opinf.lift` object or None
        Lifting transformation.
    transformer : :mod:`opinf.pre` object or None
        Preprocesser.
    basis : :mod:`opinf.basis` object or None
        Dimensionality reducer.
    ddt_estimator : :mod:`opinf.ddt` object or None
        Time derivative estimator.
        Ignored if ``model`` is not time continuous.
    """

    def __init__(
        self,
        model,
        *,
        lifter=None,
        transformer=None,
        basis=None,
        ddt_estimator=None,
    ):
        """Store each argument as an attribute."""
        if not modutils.is_nonparametric(model):
            raise TypeError("'model' must be a nonparametric model instance")
        super().__init__(model, lifter, transformer, basis, ddt_estimator)

    # Training ----------------------------------------------------------------
    def fit(
        self,
        states,
        lhs=None,
        inputs=None,
        fit_transformer: bool = True,
        fit_basis: bool = True,
    ):
        """Calibrate the model to training data.

        Parameters
        ----------
        states : list of s (n, k_i) ndarrays
            State snapshots in the original state space. Each array
            ``states[i]`` is data corresponding to a different trajectory;
            each column ``states[i][:, j]`` is one snapshot.
            If there is only one trajectory of training data (s = 1),
            ``states`` may be an (n, k) ndarray. In this case, it is assumed
            that ``lhs`` and ``inputs`` (if given) are arrays, not a sequence
            of arrays.
        lhs : list of s (n, k_i) ndarrays or None
            Left-hand side regression data. Each array ``lhs[i]`` is the data
            corresponding to parameter value ``parameters[i]``; each column
            ``lhs[i][:, j]`` corresponds to the snapshot ``states[i][:, j]``.

            - If the model is time continuous, these are the time derivatives
              of the state snapshots.
            - If the model is fully discrete, these are the "next states"
              corresponding to the state snapshots.

            If ``None``, these are estimated using :attr:`ddt_estimator`
            (time continuous) or extracted from ``states`` (fully discrete).
        inputs : list of s (m, k_i) ndarrays or None
            Input training data. Each array ``inputs[i]`` is the data
            corresponding to parameter value ``parameters[i]``; each column
            ``inputs[i][:, j]`` corresponds to the snapshot ``states[:, j]``.
            May be a two-dimensional array if :math:`m=1` (scalar input).
            Only required if one or more model operators depend on inputs.
        fit_transformer : bool
            If ``True`` (default), calibrate the preprocessing transformation
            using the ``states``. If ``False``, assume the transformer is
            already calibrated.
        fit_basis : bool
            If ``True``, calibrate the high-to-low dimensional mapping
            using the ``states``.
            If ``False``, assume the basis is already calibrated.

        Returns
        -------
        self
        """
        self._fit_and_return_training_data(
            parameters=None,
            states=states,
            lhs=lhs,
            inputs=inputs,
            fit_transformer=fit_transformer,
            fit_basis=fit_basis,
        )
        return self

    def fit_regselect_continuous(
        self,
        candidates: list,
        train_time_domains: np.ndarray,
        states: list,
        ddts: list = None,
        input_functions: list = None,
        fit_transformer: bool = True,
        fit_basis: bool = True,
        regularizer_factory=None,
        gridsearch_only: bool = False,
        test_time_length: float = 0,
        stability_margin: float = 5.0,
        test_cases: list = None,
        verbose: bool = False,
        **predict_options: dict,
    ):
        """Calibrate the time-continuous model to training data, selecting the
        regularization hyperparameter(s) that minimize the training error while
        maintaining stability over the testing regime.

        This method requires the :attr:`model` to be time-continuous (use
        :meth:`fit_regselect_discrete` for fully discrete models) and to have a
        ``solver`` of one of the following types:

        * :class:`opinf.lstsq.L2Solver`
        * :class:`opinf.lstsq.L2DecoupledSolver`
        * :class:`opinf.lstsq.TikhonovSolver`
        * :class:`opinf.lstsq.TikhonovDecoupledSolver`

        The ``model.solver.regularizer`` is repeatedly adjusted, and the model
        is recalibrated, until a best regularization is selected.
        See :cite:`mcquarrie2021combustion`.

        Parameters
        ----------
        candidates : list of regularization hyperparameters
            Regularization hyperparameters to check before carrying out a
            derivative-free optimization.
        train_time_domains : list of s (k_i,) ndarrays
            Time domain corresponding to the training states.
        states : list of s (n, k_i) ndarrays
            State snapshots in the original state space. Each array
            ``states[i]`` is data corresponding to a different trajectory;
            each column ``states[i][:, j]`` is one snapshot.
        ddts : list of s (n, k_i) ndarrays or None
            Snapshot time derivative data. Each array ``ddts[i]`` are the time
            derivatives of ``states[i]``; each column ``ddts[i][:, j]``
            corresponds to the snapshot ``states[i][:, j]``. If ``None``
            (default), these are estimated using :attr:`ddt_estimator`.
        input_functions : list of s callables or None
            Input functions mapping time to input vectors. Only required if the
            :attr:`model` takes external inputs. Each ``input_functions[i]``
            is the function corresponding to ``states[i]``, and
            ``input_functions[i](train_time_domains[i][j])`` is the input
            vector corresponding to the snapshot ``states[i][:, j]``.
        fit_transformer : bool
            If ``True`` (default), calibrate the preprocessing transformation
            using the ``states``.
            If ``False``, assume the transformer is already calibrated.
        fit_basis : bool
            If ``True`` (default), calibrate the high-to-low dimensional
            mapping using the ``states``.
            If ``False``, assume the basis is already calibrated.
        regularizer_factory : callable or None
            Function mapping regularization hyperparameters to the full
            regularizer. Specifically, ``regularizer_factory(candidates[i])``
            will be assigned to ``model.solver.regularizer`` for each ``i``.
            If ``None`` (default), set ``regularizer_factory()`` to the
            identity function.
        gridsearch_only : bool
            If ``True``, stop after checking all regularization ``candidates``
            and do not follow up with optimization.
        test_time_length : float or None
            Amount of time after the training regime in which to require model
            stability.
        stability_margin : float
            Factor by which the predicted reduced states may deviate from the
            range of the training reduced states without the trajectory being
            classified as unstable.
        test_cases : list of ContinuousRegTest objects
            Additional test cases for which the model is required to be stable.
            See :class:`opinf.utils.ContinuousRegTest`.
        verbose : bool
            If ``True``, print information during the regularization selection.
        predict_options : dict or None
            Extra arguments for :meth:`opinf.models.ContinuousModel.predict`.

        Notes
        -----
        If there is only one trajectory of training data (s = 1), ``states``
        may be provided as an (n, k) ndarray. In this case, it is assumed that
        ``ddts`` (if provided) is an (n, k) ndarray.

        The ``train_time_domains`` may be a single one-dimensional array, in
        which case it is assumed that each trajectory ``states[i]`` corresponds
        to the same time domain. Similarly, if ``input_functions`` is a single
        callable, it is assumed to be the input function for each trajectory.
        """
        return super().fit_regselect_continuous(
            candidates=candidates,
            train_time_domains=train_time_domains,
            parameters=None,
            states=states,
            ddts=ddts,
            input_functions=input_functions,
            fit_transformer=fit_transformer,
            fit_basis=fit_basis,
            regularizer_factory=regularizer_factory,
            gridsearch_only=gridsearch_only,
            test_time_length=test_time_length,
            stability_margin=stability_margin,
            test_cases=test_cases,
            verbose=verbose,
            **predict_options,
        )

    def fit_regselect_discrete(
        self,
        candidates: list,
        states: list,
        inputs: list = None,
        fit_transformer: bool = True,
        fit_basis: bool = True,
        regularizer_factory=None,
        gridsearch_only: bool = False,
        num_test_iters: int = 0,
        stability_margin: float = 5.0,
        test_cases: list = None,
        verbose: bool = False,
    ):
        """Calibrate the fully discrete model to training data, selecting the
        regularization hyperparameter(s) that minimize the training error
        while maintaining stability over the testing regime.

        This method requires the :attr:`model` to be fully discrete (use
        :meth:`fit_regselect_continuous` for time-continuous models) and to
        have a ``solver`` of one of the following types:

        * :class:`opinf.lstsq.L2Solver`
        * :class:`opinf.lstsq.L2DecoupledSolver`
        * :class:`opinf.lstsq.TikhonovSolver`
        * :class:`opinf.lstsq.TikhonovDecoupledSolver`

        The ``model.solver.regularizer`` is repeatedly adjusted, and the model
        is recalibrated, until a best regularization is selected.
        See :cite:`mcquarrie2021combustion`.

        candidates : list of regularization hyperparameters
            Regularization hyperparameters to check. If a single hyperparameter
            is given, use it as the start of an optimization-based search.
        states : list of s (n, k_i) ndarrays
            State snapshots in the original state space. Each array
            ``states[i]`` is data corresponding to a different trajectory;
            each column ``states[i][:, j]`` is one snapshot. This method
            assumes the snapshots are sequential, i.e., the model maps
            ``states[i][:, j]`` to ``states[i][:, j+1]``.
        states : list of s (r, k_i) ndarrays
            State snapshots in the reduced state space. This method assumes
            the snapshots are sequential, i.e., the model maps
            ``states[i][:, j]`` to ``states[i][:, j+1]``.
        inputs : list of s (m, k_i + num_test_iters) ndarrays
            Inputs corresponding to the training data, together with inputs
            for the testing regime. Only required if the :attr:`model` takes
            external inputs.
        regularizer_factory : callable or None
            Function mapping regularization hyperparameters to the full
            regularizer. Specifically, ``regularizer_factory(candidates[i])``
            will be assigned to ``model.solver.regularizer`` for each ``i``.
        gridsearch_only : bool
            If ``True``, stop after checking all regularization ``candidates``
            and do not follow up with optimization.
        num_test_iters : int
            Number of iterations after the training data in which to require
            model stability.
        stability_margin : float,
            Factor by which the reduced states may deviate from the range of
            the training data without being flagged as unstable.
        test_cases : list of DiscreteRegTest objects
            Additional test cases for which the model is required to be stable.
            See :class:`opinf.utils.DiscreteRegTest`.
        verbose : bool
            If ``True``, print information during the regularization selection.

        Notes
        -----
        If there is only one trajectory of training data (s = 1), ``states``
        may be provided as an (n, k) ndarray. In this case, it is assumed that
        ``inputs`` (if provided) is a single (m, k + num_test_iters) ndarray.
        """
        return super().fit_regselect_discrete(
            candidates=candidates,
            parameters=None,
            states=states,
            inputs=inputs,
            fit_transformer=fit_transformer,
            fit_basis=fit_basis,
            regularizer_factory=regularizer_factory,
            gridsearch_only=gridsearch_only,
            num_test_iters=num_test_iters,
            stability_margin=stability_margin,
            test_cases=test_cases,
            verbose=verbose,
        )

    # Evaluation --------------------------------------------------------------
    def predict(self, state0, *args, **kwargs):
        """Evaluate the reduced-order model.

        Arguments are the same as the ``predict()`` method of :attr:`model`.

        Parameters
        ----------
        state0 : (n,) ndarray
            Initial state, expressed in the original state space.
        *args : list
            Other positional arguments to the ``predict()`` method of
            :attr:`model`.
        **kwargs : dict
            Keyword arguments to the ``predict()`` method of :attr:`model`.

        Returns
        -------
        states: (n, k) ndarray
            Solution to the model, expressed in the original state space.
        """
        q0_ = self.encode(state0, fit_transformer=False, fit_basis=False)
        states = self.model.predict(q0_, *args, **kwargs)
        return self.decode(states)



# operators/_elementwise.py
"""Operators for Hadamard (element-wise) products."""

from ._base import OperatorTemplate


class ElementwiseQuadraticOperator(OperatorTemplate):
    def __init__(self, state_dimension=None):
        self.__r = None

    @property
    def state_dimension(self):
        return self.__r

    def apply(self, state, input_=None):
        if self.__r is None:
            self.__r = state.shape[0]
        return state**2

    def jacobian(self, state, input_=None):
        pass



# operators/_base.py
"""Abstract base classes for operators."""

__all__ = [
    "InputMixin",
    "OperatorTemplate",
    "OpInfOperator",
    "ParametricOperatorTemplate",
    "ParametricOpInfOperator",
]

import os
import abc
import copy
import types
import numpy as np
import scipy.linalg as la
import scipy.sparse as sparse
import matplotlib.pyplot as plt

from .. import errors, utils


class InputMixin(abc.ABC):
    r"""Mixin for operators whose ``apply()`` method acts on the input
    :math:`\u`.

    Operators that do not inherit from this Mixin do not have an
    ``input_dimension`` attribute, which indicates :math:`m = 0`.
    """

    @property
    @abc.abstractmethod
    def input_dimension(self) -> int:
        r"""Dimension :math:`m` of the input :math:`\u` that the operator
        acts on.
        """
        raise NotImplementedError  # pragma: no cover


def has_inputs(obj) -> bool:
    r"""Return ``True`` if ``obj`` is an operator object whose ``apply()``
    method acts on the ``input_`` argument, i.e.,
    :math:`\Ophat_{\ell}(\qhat,\u)` depends on :math:`\u`.
    """
    return isinstance(obj, InputMixin)


# Nonparametric operators =====================================================
class OperatorTemplate(abc.ABC):
    r"""Template for general operators :math:`\Ophat_{\ell}(\qhat,\u).`

    In this package, an "operator" is a function
    :math:`\Ophat_{\ell}: \RR^r \times \RR^m \to \RR^r` that acts on a state
    vector :math:`\qhat\in\RR^r` and (optionally) an input vector
    :math:`\u\in\RR^m`.

    Models are defined as the sum of several operators,
    for example, an :class:`opinf.models.ContinuousModel` object represents a
    system of ordinary differential equations:

    .. math::
       \ddt\qhat(t)
       = \sum_{\ell=1}^{n_\textrm{terms}}\Ophat_{\ell}(\qhat(t),\u(t)).

    Notes
    -----
    This class can be used for custom nonparametric model terms that are not
    learnable with Operator Inference.
    For parametric model terms, see :class:`ParametricOperatorTemplate`.
    For model terms that can be learned with Operator Inference, see
    :class:`OpInfOperator` or :class:`ParametricOpInfOperator`.
    """

    # Properties --------------------------------------------------------------
    @property
    @abc.abstractmethod
    def state_dimension(self) -> int:
        r"""Dimension :math:`r` of the state :math:`\qhat` that the operator
        acts on.
        """
        raise NotImplementedError  # pragma: no cover

    def __str__(self) -> str:
        """String representation: class name + dimensions."""
        out = [self.__class__.__name__]
        out.append(f"state_dimension: {self.state_dimension}")
        if has_inputs(self):
            out.append(f"input_dimension: {self.input_dimension}")
        return "\n  ".join(out)

    def __repr__(self) -> str:
        return utils.str2repr(self)

    @staticmethod
    def _str(statestr, inputstr=None):
        """String representation of the operator, used when printing out the
        structure of a model.

        Parameters
        ----------
        statestr : str
            String representation of the state, e.g., ``"q(t)"`` or ``"q_j"``.
        inputstr : str
            String representation of the input, e.g., ``"u(t)"`` or ``"u_j"``.

        Returns
        -------
        opstr : str
            String representation of the operator acting on the state/input,
            e.g., ``"Aq(t)"`` or ``"Bu(t)"`` or ``"H[q(t) ⊗ q(t)]"``.
        """
        return f"f({statestr}, {inputstr})"  # pragma: no cover

    # Evaluation --------------------------------------------------------------
    @abc.abstractmethod
    def apply(self, state: np.ndarray, input_=None) -> np.ndarray:
        """Apply the operator mapping to the given state / input.

        Parameters
        ----------
        state : (r,) or (r, k) ndarray
            State vector or matrix of state vectors.
        input_ : (m,) or (m, k) ndarray or None
            Input vector or matrix of input vectors.

        Returns
        -------
        out : (r,) or (r, k) ndarray
            Application of the operator to the state / input, with the same
            number of dimensions as ``state`` and (if provided) ``input_``.
        """
        raise NotImplementedError  # pragma: no cover

    def jacobian(self, state: np.ndarray, input_=None) -> np.ndarray:
        r"""Construct the state Jacobian of the operator.

        If :math:`[\![\q]\!]_{i}` denotes the :math:`i`-th entry of a vector
        :math:`\q`, then the :math:`(i,j)`-th entry of the state Jacobian is
        given by

        .. math::
           [\![\ddqhat\Ophat(\qhat,\u)]\!]_{i,j}
           = \frac{\partial}{\partial[\![\qhat]\!]_j}
           [\![\Ophat(\qhat,\u)]\!]_i.

        Parameters
        ----------
        state : (r,) ndarray
            State vector.
        input_ : (m,) ndarray or float or None
            Input vector.

        Returns
        -------
        jac : (r, r) ndarray
            State Jacobian.
        """
        raise NotImplementedError  # pragma: no cover

    # Dimensionality reduction ------------------------------------------------
    def galerkin(self, Vr: np.ndarray, Wr=None):
        r"""Get the (Petrov-)Galerkin projection of this operator.

        Consider an operator :math:`\Op(\q,\u)`, where :math:`\q\in\RR^n`
        is the state and :math:`\u\in\RR^m` is the input.
        Given a *trial basis* :math:`\Vr\in\RR^{n\times r}` and a *test basis*
        :math:`\Wr\in\RR^{n\times r}`, the Petrov-Galerkin projection of
        :math:`\Op` is the operator :math:`\Ophat:\RR^r\times\RR^m\to\RR^r`
        defined by

        .. math::
           \Ophat(\qhat, \u) = (\Wr\trp\Vr)^{-1}\Wr\trp\Op(\Vr\qhat, \u)

        where :math:`\qhat\in\RR^n` approximates the original state via
        :math:`\q \approx \Vr\qhat`.

        Parameters
        ----------
        Vr : (n, r) ndarray
            Basis for the trial space :math:`\Vr`.
        Wr : (n, r) ndarray or None
            Basis for the test space :math:`\Wr`.
            If ``None`` (default), use ``Vr`` as the test basis.

        Returns
        -------
        op : :class:`OperatorTemplate`
            New operator object whose ``state_dimension``
            attribute equals ``r``. If this operator acts on inputs, the
            ``input_dimension`` attribute of the new operator should be
            ``self.input_dimension``.
        """
        raise NotImplementedError  # pragma: no cover

    # Model persistence -------------------------------------------------------
    def copy(self):
        """Return a copy of the operator using :func:`copy.deepcopy()`."""
        return copy.deepcopy(self)  # pragma: no cover

    def save(self, savefile: str, overwrite: bool = False) -> None:
        """Save the operator to an HDF5 file.

        Parameters
        ----------
        savefile : str
            Path of the file to save the basis in.
        overwrite : bool
            If ``True``, overwrite the file if it already exists. If ``False``
            (default), raise a ``FileExistsError`` if the file already exists.
        """
        raise NotImplementedError  # pragma: no cover

    @classmethod
    def load(cls, loadfile: str):
        """Load an operator from an HDF5 file.

        Parameters
        ----------
        loadfile : str
            Path to the file where the operator was stored via :meth:`save()`.
        """
        raise NotImplementedError  # pragma: no cover

    # Verification ------------------------------------------------------------
    def verify(
        self,
        plot: bool = False,
        *,
        k: int = 10,
        fdifftol: float = 1e-5,
        ntests: int = 4,
    ) -> None:
        """Verify consistency between dimension properties and required
        methods.

        This method verifies :meth:`apply()` and, if implemented,
        :meth:`jacobian()`, :meth:`galerkin()`, :meth:`copy()`,
        :meth:`save()`, and :meth:`load()`.

        Parameters
        ----------
        plot : bool
            If ``True``, plot the relative errors of the finite difference
            check for :meth:`jacobian()` as a function of the perturbation
            size.
            If ``False`` (default), print a report of the relative errors.
            Nothing is plotted or printed if :meth:`jacobian()` is not
            implemented.

        Notes
        -----
        This method does **not** verify the correctness of :meth:`apply()`,
        only that it returns an output with the expected shape. However,
        if :meth:`jacobian()` is implemented, a finite difference check is
        applied to check that :meth:`apply()` and :meth:`jacobian()` are
        consistent.
        """
        # Verify dimensions exist and are valid.
        if (
            not isinstance((r := self.state_dimension), int) or r <= 0
        ):  # pragma: no cover
            raise errors.VerificationError(
                "state_dimension must be a positive integer "
                f"(current value: {repr(r)}, of type '{type(r).__name__}')"
            )

        if hasinputs := has_inputs(self):
            if (
                not isinstance((m := self.input_dimension), int) or m <= 0
            ):  # pragma: no cover
                raise errors.VerificationError(
                    "input_dimension must be a positive integer "
                    f"(current value: {repr(m)}, of type '{type(r).__name__}')"
                )
        else:
            m = 0

        # Verify apply() - - - - - - - - - - - - - - - - - - - - - - - - - - -
        Q = np.random.random((r, k))
        q = Q[:, 0]
        U, u = None, None
        if hasinputs:
            U = np.random.random((m, k))
            u = U[:, 0]

        out = self.apply(q, u)
        if not isinstance(out, np.ndarray) or out.shape != (
            r,
        ):  # pragma: no cover
            _message = [
                "apply(q, u) must return array of shape (state_dimension,)",
                "when q.shape = (state_dimension,)",
                "and u = None",
            ]
            if hasinputs:
                _message[-1] = "and u.shape = (input_dimension,)"
            raise errors.VerificationError(" ".join(_message))

        out = self.apply(Q, U)
        if not isinstance(out, np.ndarray) or out.shape != (
            r,
            k,
        ):  # pragma: no cover
            _message = [
                "apply(Q, U) must return array of shape (state_dimension, k)",
                "when Q.shape = (state_dimension, k)",
                "and U = None",
            ]
            if hasinputs:
                _message[-1] = "and U.shape = (input_dimension, k)"
            raise errors.VerificationError(" ".join(_message))

        # Report successes.
        def _report_isconsistent(method):
            _message = f"{method} is consistent with state_dimension"
            if hasinputs:
                _message += " and input_dimension"
            print(_message)

        _report_isconsistent("apply()")

        # Verify jacobian() - - - - - - - - - - - - - - - - - - - - - - - - - -
        def _gradient(f, x, h=1e-8):
            """Estimate the Jacobian of f:R^n -> R^m at x using perturbations
            of magnitude h.
            """
            E = np.eye((n := x.size))
            return np.array([(f(x + h * E[i]) - f(x)) / h for i in range(n)]).T

        def _finite_difference_check(f, df, x, hs=None):
            """Compare analytical and numerical derivatives."""
            dfx = df(x)
            if hs is None:
                hs = np.logspace(-10, -1, 10)[::-1]
            return hs, np.array(
                [la.norm(_gradient(f, x, h) - dfx) / la.norm(dfx) for h in hs]
            )

        try:
            out = self.jacobian(q, u)
        except NotImplementedError:  # pragma: no cover
            print("jacobian() not implemented")
        else:
            if np.isscalar(out) and out == 0:  # pragma: no cover
                print("jacobian() = 0")
            elif not isinstance(out, np.ndarray) or out.shape != (
                r,
                r,
            ):  # pragma: no cover
                _message = [
                    "jacobian(q, u) must return array",
                    "of shape (state_dimension, state_dimension)",
                    "when q.shape = (state_dimension,)",
                    "and u = None",
                ]
                if hasinputs:
                    _message[-1] = "and u.shape = (input_dimension,)"
                raise errors.VerificationError(" ".join(_message))
            else:
                _report_isconsistent("jacobian()")

                # Finite difference check.
                hs, diffs = _finite_difference_check(
                    lambda x: self.apply(x, u),
                    lambda x: self.jacobian(x, u),
                    np.random.standard_normal(r),
                )
                if plot:  # pragma: no cover
                    plt.loglog(hs, diffs, ".-", markersize=5, linewidth=0.5)
                else:
                    print(
                        "jacobian() finite difference relative errors",
                        "  ------------------------------------------",
                        sep="\n",
                    )
                    for h, err in zip(hs, diffs):
                        print(f"  h = {h:.2e}\terror = {err:.4e}")
                if np.min(diffs) > fdifftol:  # pragma: no cover
                    raise errors.VerificationError(
                        "jacobian() finite difference check failed"
                    )

        # Verify galerkin() - - - - - - - - - - - - - - - - - - - - - - - - - -
        def _orth(n, k):
            """Get an n x k matrix with orthonormal columns."""
            return la.qr(np.random.standard_normal((n, k)), mode="economic")[0]

        if r > 1:
            rnew = r // 2
            Vr = _orth(r, rnew)
            Wr = _orth(r, rnew)
            try:
                out = self.galerkin(Vr, Wr)
            except NotImplementedError:  # pragma: no cover
                print("galerkin() not implemented")
            else:
                if not isinstance(out, OperatorTemplate):  # pragma: no cover
                    raise errors.VerificationError(
                        "galerkin() must return object "
                        "whose class inherits from OperatorTemplate"
                    )
                if out.state_dimension != rnew:  # pragma: no cover
                    raise errors.VerificationError(
                        "galerkin(Vr, Wr).state_dimension != Vr.shape[1]"
                    )
                if hasinputs and out.input_dimension != m:  # pragma: no cover
                    raise errors.VerificationError(
                        "self.galerkin(Vr, Wr).input_dimension "
                        "!= self.input_dimension"
                    )
                WrTVr_LU = la.lu_factor(Wr.T @ Vr)
                for _ in range(ntests):
                    qr = np.random.random(rnew)
                    full = la.lu_solve(WrTVr_LU, Wr.T @ self.apply(Vr @ qr, u))
                    reduced = out.apply(qr, u)
                    if not np.allclose(reduced, full):  # pragma: no cover
                        raise errors.VerificationError(
                            "op2.apply(qr, u) != "
                            "inv(Wr.T @ Vr) @ Wr.T @ self.apply(Vr @ qr, u) "
                            "where op2 = self.galerkin(Vr, Wr)"
                        )
                out = self.galerkin(Vr)
                for _ in range(ntests):
                    qr = np.random.random(rnew)
                    full = Vr.T @ self.apply(Vr @ qr, u)
                    reduced = out.apply(qr, u)
                    if not np.allclose(reduced, full):  # pragma: no cover
                        raise errors.VerificationError(
                            "op2.apply(qr, u) != "
                            "Vr.T @ self.apply(Vr @ qr, u) "
                            "where op2 = self.galerkin(Vr) and Vr.T @ Vr = I"
                        )
                print("galerkin() is consistent with apply()")
        else:  # pragma: no cover
            print("cannot test galerkin() when state_dimension = 1")

        # Verify copy() - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        out = self.copy()
        if out is self:  # pragma: no cover
            raise errors.VerificationError("self.copy() is self")
        if out.__class__ is not self.__class__:  # pragma: no cover
            raise errors.VerificationError(
                "type(self.copy()) is not type(self)"
            )
        if out.state_dimension != r:  # pragma: no cover
            raise errors.VerificationError(
                "self.copy().state_dimension != self.state_dimension"
            )
        if hasinputs and out.input_dimension != m:  # pragma: no cover
            raise errors.VerificationError(
                "self.copy().input_dimension != self.input_dimension"
            )
        _report_isconsistent("copy()")
        for _ in range(ntests):
            q = np.random.random(r)
            if not np.allclose(
                out.apply(q, u), self.apply(q, u)
            ):  # pragma: no cover
                raise errors.VerificationError(
                    "self.copy().apply() not consistent with self.apply()"
                )
        print("copy() preserves the results of apply()")

        # Verify save()/load() - - - - - - - - - - - - - - - - - - - - - - - -
        tempfile = "_operatorverification.h5"
        try:
            self.save(tempfile)
            out = self.load(tempfile)
        except NotImplementedError:  # pragma: no cover
            print("save() and/or load() not implemented")
        else:
            if out.__class__ is not self.__class__:  # pragma: no cover
                raise errors.VerificationError(
                    "save()/load() does not preserve object type"
                )
            if out.state_dimension != r:  # pragma: no cover
                raise errors.VerificationError(
                    "save()/load() does not preserve state_dimension"
                )
            if hasinputs and out.input_dimension != m:  # pragma: no cover
                raise errors.VerificationError(
                    "save()/load() does not preserve input_dimension"
                )
            _report_isconsistent("save()/load()")
            for _ in range(ntests):
                q = np.random.random(r)
                if not np.allclose(
                    out.apply(q, u), self.apply(q, u)
                ):  # pragma: no cover
                    raise errors.VerificationError(
                        "save()/load() does not preserve the result of apply()"
                    )
            print("save()/load() preserves the results of apply()")
        finally:
            if os.path.isfile(tempfile):  # pragma: no cover
                os.remove(tempfile)


def is_nonparametric(obj) -> bool:
    """Return ``True`` if ``obj`` is a nonparametric operator object."""
    return isinstance(obj, OperatorTemplate)


class OpInfOperator(OperatorTemplate):
    r"""Template for nonparametric operators that can be calibrated through
    Operator Inference, i.e.,
    :math:`\Ophat_{\ell}(\qhat, \u) = \Ohat_{\ell}\d(\qhat, \u)`.

    In this package, an "operator" is a function
    :math:`\Ophat_{\ell}: \RR^r \times \RR^m \to \RR^r` that acts on a state
    vector :math:`\qhat\in\RR^r` and (optionally) an input vector
    :math:`\u\in\RR^m`.

    Models are defined as the sum of several operators,
    for example, an :class:`opinf.models.ContinuousModel` object represents a
    system of ordinary differential equations:

    .. math::
       \ddt\qhat(t)
       = \sum_{\ell=1}^{n_\textrm{terms}}\Ophat_{\ell}(\qhat(t),\u(t)).

    Operator Inference calibrates operators that can be written as the product
    of a matrix and some known (possibly nonlinear) function of the state
    and/or input:

    .. math::
       \Ophat_{\ell}(\qhat, \u)
       = \Ohat_{\ell}\d(\qhat, \u),

    where :math:`\Ohat_{\ell}\in\RR^{r\times d}` is a constant matrix, called
    the *operator entries*, and :math:`\d_{\ell}:\RR^r\times\RR^m\to\RR^d`.

    Notes
    -----
    * To define operators with a more general structure than
      :math:`\Ohat_{\ell}\d(\qhat, \u),` see :class:`OperatorTemplate`.
    * If the operator entries :math:`\Ohat_{\ell}` depend on one or more
      external parameters, it is called a *parametric operator*.
      See :class:`ParametricOpInfOperator`.
    """

    # Initialization ----------------------------------------------------------
    def __init__(self, entries=None):
        """Initialize an empty operator."""
        self._clear()
        if entries is not None:
            self.set_entries(entries)

    def _clear(self):
        """Delete operator ``entries`` and related attributes."""
        self.__entries = None

    @staticmethod
    def _validate_entries(entries):
        """Ensure argument is a NumPy array and screen for NaN, Inf entries."""
        if sparse.issparse(entries):
            return
        if not isinstance(entries, np.ndarray):
            raise TypeError(
                "operator entries must be NumPy or scipy.sparse array"
            )
        if np.any(np.isnan(entries)):
            raise ValueError("operator entries must not be NaN")
        elif np.any(np.isinf(entries)):
            raise ValueError("operator entries must not be Inf")

    def set_entries(self, entries) -> None:
        """Set the :attr:`entries` attribute."""
        self.__entries = entries

    # Properties --------------------------------------------------------------
    @property
    def entries(self) -> np.ndarray:
        r"""Discrete representation of the operator,
        the matrix :math:`\Ohat`.
        """
        return self.__entries

    @entries.setter
    def entries(self, entries):
        """Set the ``entries`` attribute."""
        self.set_entries(entries)

    @entries.deleter
    def entries(self):
        """Reset the ``entries`` attribute."""
        self._clear()

    @property
    def shape(self) -> tuple:
        """Shape of the operator matrix."""
        return None if self.entries is None else self.entries.shape

    @property
    def state_dimension(self) -> int:
        r"""Dimension :math:`r` of the state :math:`\qhat` that the operator
        acts on.
        """
        return None if self.entries is None else self.entries.shape[0]

    # Magic methods -----------------------------------------------------------
    def __getitem__(self, key):
        """Slice into the entries of the operator."""
        return None if self.entries is None else self.entries[key]

    def __eq__(self, other):
        """Two OpInf operators are equal if they are of the same class
        and have the same ``entries`` array.
        """
        if not isinstance(other, self.__class__):
            return False
        if (self.entries is None and other.entries is not None) or (
            self.entries is not None and other.entries is None
        ):
            return False
        if self.entries is not None:
            if self.shape != other.shape:
                return False
            return np.all(self.entries == other.entries)
        return True

    def __add__(self, other):
        """Nonparametric operators are linear in their entries."""
        if (ocls := other.__class__) is not (scls := self.__class__):
            raise TypeError(
                f"can't add object of type '{ocls.__name__}' "
                f"to object of type '{scls.__name__}'"
            )
        return scls(self.entries + other.entries)

    def __str__(self):
        out = OperatorTemplate.__str__(self)
        return out + f"\n  entries.shape:   {self.shape}"

    # Evaluation --------------------------------------------------------------
    @utils.requires("entries")
    def jacobian(self, state, input_=None) -> np.ndarray:  # pragma: no cover
        r"""Construct the state Jacobian of the operator.

        If :math:`[\![\q]\!]_{i}` denotes the :math:`i`-th entry of a vector
        :math:`\q`, then the :math:`(i,j)`-th entry of the state Jacobian is
        given by

        .. math::
           [\![\ddqhat\Ophat(\qhat,\u)]\!]_{i,j}
           = \frac{\partial}{\partial[\![\qhat]\!]_j}
           [\![\Ophat(\qhat,\u)]\!]_i.

        If a child class does not implement this method, it is assumed that
        the Jacobian is zero (i.e., the operator does not act on the state).

        Parameters
        ----------
        state : (r,) ndarray
            State vector.
        input_ : (m,) ndarray or float or None
            Input vector.

        Returns
        -------
        jac : (r, r) ndarray
            State Jacobian.
        """
        return 0

    # Dimensionality reduction ------------------------------------------------
    def _galerkin(self, Vr, Wr, func):
        r"""Get the (Petrov-)Galerkin projection of this operator.

        Subclasses may implement this function as follows:

        .. code-block:: python

           @utils.requires("entries")
           def galerkin(self, Vr, Wr=None):
               '''Docstring'''
               return self._galerkin(Vr, Wr, lambda A, V: <TODO>)

        Parameters
        ----------
        Vr : (n, r) ndarray
            Basis for the trial space.
        Wr : (n, r) ndarray or None
            Basis for the test space. If ``None``, defaults to ``Vr``.
        func : callable
            Function that accepts the operator ``entries`` and the trial
            basis ``Vr`` and computes the right side of the projection,
            i.e, the result of substituting :math:`\q` with :math:`\Vr\qhat`.
            For example, for linear operator :math:`\Op_\ell(\q) = \A\q`,
            ``func`` should return :math:`\A\Vr`.

        Returns
        -------
        op : operator
            New object of the same class as ``self``.
        """
        if Wr is None:
            Wr = Vr
        n, r = Wr.shape
        if self.entries.shape[0] != n:  # pragma: no cover
            raise errors.DimensionalityError("basis and operator not aligned")
        if Vr.shape[1] != r:  # pragma: no cover
            raise errors.DimensionalityError(
                "trial and test bases not aligned"
            )

        entries = Wr.T @ func(self.entries, Vr)
        if not np.allclose((WrTVr := Wr.T @ Vr), np.eye(r)):
            entries = la.solve(WrTVr, entries)
        return self.__class__(entries)

    # Operator inference ------------------------------------------------------
    @staticmethod
    @abc.abstractmethod
    def operator_dimension(r: int, m: int = None) -> int:
        r"""Column dimension of the operator entries.

        Child classes should decorate this method with ``@staticmethod``.

        Parameters
        ----------
        r : int
            State dimension.
        m : int or None
            Input dimension.

        Returns
        -------
        d : int
            Number of columns in the operator entries matrix.
            This is also the number of rows in the data matrix block.
        """
        raise NotImplementedError  # pragma: no cover

    @staticmethod
    @abc.abstractmethod
    def datablock(states: np.ndarray, inputs=None) -> np.ndarray:
        r"""Construct the data matrix block corresponding to the operator.

        For a nonparametric operator
        :math:`\Ophat_{\ell}(\qhat,\u) = \Ohat_{\ell}\d_{\ell}(\qhat, \u)`,
        the data matrix block corresponding to the data pairs
        :math:`\{(\qhat_j,\u_j)\}_{j=0}^{k-1}` is the matrix

        .. math::
           \D\trp = \left[\begin{array}{c|c|c|c}
           & & & \\
           \d_{\ell}(\qhat_0,\u_0) & \d_{\ell}(\qhat_1,\u_1)
           & \cdots &
           \\d_{\ell}(\qhat_{k-1},\u_{k-1})
           \\ & & &
           \end{array}\right]
           \in \RR^{d \times k}.

        Here, ``states`` is the snapshot matrix
        :math:`[~\qhat_0~~\cdots~~\qhat_{k-1}~]`
        and ``inputs`` is the (optional) input matrix
        :math:`[~\u_0~~\cdots~~\u_{k-1}~]`.

        Child classes should decorate this method with ``@staticmethod``.

        Parameters
        ----------
        states : (r, k) or (k,) ndarray
            State vectors. Each column is a single state vector.
            If one dimensional, it is assumed that :math:`r = 1`.
        inputs : (m, k) or (k,) ndarray or None
            Input vectors. Each column is a single input vector.
            If one dimensional, it is assumed that :math:`m = 1`.

        Returns
        -------
        block : (d, k) or (d,) ndarray
            Data matrix block. Here, :math:`d` is ``entries.shape[1]``.
        """
        raise NotImplementedError  # pragma: no cover

    # Model persistence -------------------------------------------------------
    def copy(self):
        """Return a copy of the operator."""
        entries = self.entries.copy() if self.entries is not None else None
        return self.__class__(entries)

    def save(self, savefile: str, overwrite: bool = False) -> None:
        """Save the operator to an HDF5 file.

        Parameters
        ----------
        savefile : str
            Path of the file to save the basis in.
        overwrite : bool
            If ``True``, overwrite the file if it already exists. If ``False``
            (default), raise a ``FileExistsError`` if the file already exists.
        """
        with utils.hdf5_savehandle(savefile, overwrite) as hf:
            meta = hf.create_dataset("meta", shape=(0,))
            meta.attrs["class"] = self.__class__.__name__
            if self.entries is not None:
                hf.create_dataset("entries", data=self.entries)

    @classmethod
    def load(cls, loadfile: str):
        """Load an operator from an HDF5 file.

        Parameters
        ----------
        loadfile : str
            Path to the file where the operator was stored via :meth:`save()`.
        """
        with utils.hdf5_loadhandle(loadfile) as hf:
            if (
                ClassName := hf["meta"].attrs["class"]
            ) != cls.__name__:  # pragma: no cover
                raise TypeError(
                    f"file '{loadfile}' contains '{ClassName}' "
                    f"object, use '{ClassName}.load()"
                )
            return cls(hf["entries"][:] if "entries" in hf else None)

    # Verification ------------------------------------------------------------
    def verify(
        self,
        plot: bool = False,
        *,
        k: int = 10,
        ntests: int = 4,
        r: int = 6,
        m: int = 3,
    ) -> None:
        """Verify consistency between dimension properties and required
        methods.

        This method verifies :meth:`apply()` and, if implemented,
        :meth:`jacobian()`, :meth:`galerkin()`, :meth:`copy()`,
        :meth:`save()`, and :meth:`load()`.
        If the :attr:`entries` are not set, no checks are made.

        Parameters
        ----------
        plot : bool
            If ``True``, plot the relative errors of the finite difference
            check for :meth:`jacobian()` as a function of the perturbation
            size.
            If ``False`` (default), print a report of the relative errors.
            Nothing is plotted or printed if :meth:`jacobian()` is not
            implemented.

        Notes
        -----
        This method does **not** verify the correctness of :meth:`apply()`,
        only that it returns an output with the expected shape. However,
        if :meth:`jacobian()` is implemented, a finite difference check is
        applied to check that :meth:`apply()` and :meth:`jacobian()` are
        consistent.
        """
        # Verify operator_dimension() - - - - - - - - - - - - - - - - - - - - -
        if not isinstance(
            self.operator_dimension, types.FunctionType
        ):  # pragma: no cover
            raise errors.VerificationError(
                "operator_dimension() must have @staticmethod decorator"
            )
        d = self.operator_dimension(r, m)
        if not isinstance(d, int) or d <= 0:  # pragma: no cover
            raise errors.VerificationError(
                "operator_dimension() must return a positive integer"
            )

        # Verify datablock() - - - - - - - - - - - - - - - - - - - - - - - - -
        if not isinstance(
            self.datablock, types.FunctionType
        ):  # pragma: no cover
            raise errors.VerificationError(
                "datablock() must have @staticmethod decorator"
            )
        Dt = self.datablock(np.random.random((r, k)), np.random.random((m, k)))
        if not isinstance(Dt, np.ndarray) or Dt.ndim != 2:  # pragma: no cover
            raise errors.VerificationError(
                "datablock() must return a two-dimensional array"
            )
        if Dt.shape != (d, k):  # pragma: no cover
            raise errors.VerificationError(
                "datablock().shape[0] != operator_dimension()"
            )
        print("operator_dimension() is consistent with datablock()")

        # Verify instance methods - - - - - - - - - - - - - - - - - - - - - - -
        if self.entries is None:  # pragma: no cover
            return print("cannot verify apply() when entries=None")

        OperatorTemplate.verify(self, plot=plot, k=k, ntests=ntests)


# Parametric operators ========================================================
class ParametricOperatorTemplate(abc.ABC):
    r"""Template for operators that depend on external parameters,
    :math:`\Ophat_{\ell}(\qhat,\u;\bfmu).`

    In this package, a parametric "operator" is a function
    :math:`\Ophat_{\ell}: \RR^r \times \RR^m \times \RR^p \to \RR^r` that acts
    on a state vector :math:`\qhat\in\RR^r`, an (optional) input vector
    :math:`\u\in\RR^m`, and a parameter vector :math:`\bfmu\in\RR^p`.

    Parametric models are defined as the sum of several operators, at least
    one of which is parametric.
    For example, a system of ODEs:

    .. math::
       \ddt\qhat(t;\bfmu)
       = \sum_{\ell=1}^{n_\textrm{terms}}\Ophat_{\ell}(\qhat(t),\u(t);\bfmu).

    Notes
    -----
    This class can be used for custom nonparametric model terms that are not
    learnable with Operator Inference.
    For nonparametric model terms, see :class:`OperatorTemplate`.
    For model terms that can be learned with Operator Inference, see
    :class:`OpInfOperator` or :class:`ParametricOpInfOperator`.
    """

    # Nonparametric operator class that this parametric operator evaluates to.
    _OperatorClass = NotImplemented

    # Properties --------------------------------------------------------------
    @property
    @abc.abstractmethod
    def state_dimension(self) -> int:
        r"""Dimension :math:`r` of the state :math:`\qhat` that the operator
        acts on.
        """
        raise NotImplementedError  # pragma: no cover

    @property
    @abc.abstractmethod
    def parameter_dimension(self) -> int:
        r"""Dimension :math:`p` of the parameter vector :math:`\bfmu` that the
        operator matrix depends on.
        """
        raise NotImplementedError  # pragma: no cover

    def __str__(self) -> str:
        """String representation: class name, dimensions, evaluation type."""
        out = [self.__class__.__name__]
        out.append(f"state_dimension:     {self.state_dimension}")
        if has_inputs(self):
            out.append(f"input_dimension:     {self.input_dimension}")
        out.append(f"parameter_dimension: {self.parameter_dimension}")
        out.append(f"evaluate(parameter) -> {self._OperatorClass.__name__}")
        return "\n  ".join(out)

    def __repr__(self) -> str:
        return utils.str2repr(self)

    # Evaluation --------------------------------------------------------------
    def _check_parametervalue_dimension(self, parameter):
        """Ensure a new parameter value has the expected shape."""
        if (pdim := self.parameter_dimension) is None:
            raise RuntimeError("parameter_dimension not set")
        if np.atleast_1d(parameter).shape[0] != pdim:
            raise ValueError(f"expected parameter of shape ({pdim:d},)")

    @abc.abstractmethod
    def evaluate(self, parameter):
        r"""Evaluate the operator at the given parameter value,
        resulting in a nonparametric operator.

        Parameters
        ----------
        parameter : (p,) ndarray or float
            Parameter value :math:`\bfmu` at which to evalute the operator.

        Returns
        -------
        op : nonparametric :mod:`opinf.operators` operator
            Nonparametric operator corresponding to the parameter value.
            This should be an instance of :class:`OperatorTemplate` (or
            a class that inherits from it).
        """
        raise NotImplementedError  # pragma: no cover

    def apply(self, parameter, state, input_=None):
        r"""Apply the operator to the given state and input
        at the specified parameter value, :math:`\Ophat_\ell(\qhat,\u;\bfmu)`.

        Parameters
        ----------
        parameter : (p,) ndarray or float
            Parameter value.
        state : (r,) ndarray
            State vector.
        input_ : (m,) ndarray or float or None
            Input vector.

        Returns
        -------
        (r,) ndarray

        Notes
        -----
        For repeated calls with the same parameter value, use
        :meth:`evaluate()` to first get the nonparametric operator
        corresponding to the parameter value.

        .. code-block::

           # Instead of this...
           >>> values = [parametric_operator.apply(parameter, q, u)
           ...           for q, u in zip(list_of_states, list_of_inputs)]
           # ...it is faster to do this.
           >>> operator_at_parameter = parametric_operator.evaluate(parameter)
           >>> values = [operator_at_parameter.apply(q, u)
           ...           for q, u in zip(list_of_states, list_of_inputs)]
        """
        return self.evaluate(parameter).apply(state, input_)

    def jacobian(self, parameter, state, input_=None):
        r"""Construct the state Jacobian of the operator,
        :math:`\ddqhat\Ophat_\ell(\qhat,\u;\bfmu)`.

        If :math:`[\![\q]\!]_{i}` denotes the entry :math:`i` of a vector
        :math:`\q`, then the entries of the state Jacobian are given by

        .. math::
           [\![\ddqhat\Ophat_\ell(\qhat,\u;\bfmu)]\!]_{i,j}
           = \frac{\partial}{\partial[\![\qhat]\!]_j}
           [\![\Ophat_\ell(\qhat,\u;\bfmu)]\!]_i.

        Parameters
        ----------
        parameter : (p,) ndarray or float
            Parameter value.
        state : (r,) ndarray
            State vector.
        input_ : (m,) ndarray or float or None
            Input vector.

        Returns
        -------
        jac : (r, r) ndarray
            State Jacobian.

        Notes
        -----
        For repeated calls with the same parameter value, use
        :meth:`evaluate()` to first get the nonparametric operator
        corresponding to the parameter value.

        .. code-block::

           # Instead of this...
           >>> values = [parametric_operator.jacobian(parameter, q, u)
           ...           for q, u in zip(list_of_states, list_of_inputs)]
           # ...it is faster to do this.
           >>> operator_at_parameter = parametric_operator.evaluate(parameter)
           >>> values = [operator_at_parameter.jacobian(q, u)
           ...           for q, u in zip(list_of_states, list_of_inputs)]
        """
        return self.evaluate(parameter).jacobian(state, input_)

    # Dimensionality reduction ------------------------------------------------
    def galerkin(self, Vr, Wr=None):
        r"""Get the (Petrov-)Galerkin projection of this operator.

        Consider an operator :math:`\Op(\q,\u)`, where :math:`\q\in\RR^n`
        is the state and :math:`\u\in\RR^m` is the input.
        Given a *trial basis* :math:`\Vr\in\RR^{n\times r}` and a *test basis*
        :math:`\Wr\in\RR^{n\times r}`, the Petrov-Galerkin projection of
        :math:`\Op` is the operator :math:`\Ophat:\RR^r\times\RR^m\to\RR^r`
        defined by

        .. math::
           \Ophat(\qhat, \u) = (\Wr\trp\Vr)^{-1}\Wr\trp\Op(\Vr\qhat, \u)

        where :math:`\qhat\in\RR^n` approximates the original state via
        :math:`\q \approx \Vr\qhat`.

        Parameters
        ----------
        Vr : (n, r) ndarray
            Basis for the trial space.
        Wr : (n, r) ndarray or None
            Basis for the test space. If ``None``, defaults to ``Vr``.

        Returns
        -------
        op : operator
            New object of the same class as ``self`` whose ``state_dimension``
            attribute equals ``r``. If this operator acts on inputs, the
            ``input_dimension`` attribute of the new operator should be
            ``self.input_dimension``.
        """
        raise NotImplementedError  # pragma: no cover

    # Model persistence -------------------------------------------------------
    def copy(self):
        """Return a copy of the operator."""
        return copy.deepcopy(self)  # pragma: no cover

    def save(self, savefile: str, overwrite: bool = False) -> None:
        """Save the operator to an HDF5 file.

        Parameters
        ----------
        savefile : str
            Path of the file to save the basis in.
        overwrite : bool
            If ``True``, overwrite the file if it already exists. If ``False``
            (default), raise a ``FileExistsError`` if the file already exists.
        """
        raise NotImplementedError  # pragma: no cover

    @classmethod
    def load(cls, loadfile: str):
        """Load an operator from an HDF5 file.

        Parameters
        ----------
        loadfile : str
            Path to the file where the operator was stored via :meth:`save()`.
        """
        raise NotImplementedError  # pragma: no cover

    # Verification ------------------------------------------------------------
    def verify(self, testparam=None):
        """Verify dimension attributes and :meth:`evaluate()`.

        Parameters
        ----------
        testparam : (p,) ndarray or None
            Test parameter at which to evaluate the operator.
            If ``None`` (default), draw test parameter entries from the
            standard Normal distribution.
        """
        # Check the _OperatorClass.
        if not issubclass(
            self._OperatorClass, OperatorTemplate
        ):  # pragma: no cover
            raise errors.VerificationError(
                "_OperatorClass must be a nonparametric operator type"
            )

        # Verify dimensions exist and are valid.
        if (
            not isinstance((r := self.state_dimension), int) or r <= 0
        ):  # pragma: no cover
            raise errors.VerificationError(
                "state_dimension must be a positive integer "
                f"(current value: {repr(r)}, of type '{type(r).__name__}')"
            )

        if hasinputs := has_inputs(self):
            if (
                not isinstance((m := self.input_dimension), int) or m <= 0
            ):  # pragma: no cover
                raise errors.VerificationError(
                    "input_dimension must be a positive integer "
                    f"(current value: {repr(m)}, of type '{type(r).__name__}')"
                )
        else:
            m = 0

        # Get a test parameter.
        if testparam is None:
            testparam = np.random.standard_normal(self.parameter_dimension)
        if np.shape(testparam) != (
            self.parameter_dimension,
        ):  # pragma: no cover
            raise ValueError("testparam.shape != (parameter_dimension,)")

        # Evaluate the operator at the test parameter.
        op_evaluated = self.evaluate(testparam)
        if not isinstance(
            op_evaluated, self._OperatorClass
        ):  # pragma: no cover
            raise errors.VerificationError(
                "evaluate() must return instance of type _OperatorClass"
            )
        if not is_nonparametric(op_evaluated):  # pragma: no cover
            raise errors.VerificationError(
                "_OperatorClass must be a nonparametric operator type"
            )
        if (
            op_evaluated.state_dimension != self.state_dimension
        ):  # pragma: no cover
            raise errors.VerificationError(
                "result of evaluate() does not retain the state_dimension"
            )
        if hasinputs:
            if not has_inputs(op_evaluated):  # pragma: no cover
                raise errors.VerificationError(
                    "result of evaluate() should depend on inputs"
                )
            if op_evaluated.input_dimension != m:  # pragma: no cover
                raise errors.VerificationError(
                    "result of evaluate() does not retain the input_dimension"
                )
        else:
            if has_inputs(op_evaluated):  # pragma: no cover
                raise errors.VerificationError(
                    "result of evaluate() should not depend on inputs"
                )


def is_parametric(obj) -> bool:
    """Return ``True`` if ``obj`` is a parametric operator object."""
    return isinstance(obj, ParametricOperatorTemplate)


class ParametricOpInfOperator(ParametricOperatorTemplate):
    r"""Template for operators that depend on external parameters, and which
    can be calibrated through operator inference, i.e.,
    :math:`\Ophat_\ell(\qhat,\u;\bfmu) = \Ohat_\ell(\bfmu)\d_\ell(\qhat,\u)`.
    """

    # Initialization ----------------------------------------------------------
    def __init__(self):
        """Initialize the parameter_dimension."""
        self._clear()

    def _clear(self) -> None:
        """Reset the operator to its post-constructor state."""
        self.__p = None
        self.__entries = None

    def _set_parameter_dimension_from_values(self, parameters) -> None:
        """Extract and save the dimension of the parameter space from a set of
        one or more parameter values.

        Parameters
        ----------
        parameters : (s, p) or (s,) ndarray
            Parameter value(s).
        """
        if (dim := len(shape := np.shape(parameters))) == 1:
            self.__p = 1
        elif dim == 2:
            self.__p = shape[1]
        else:
            raise ValueError("parameter values must be scalars or 1D arrays")

    # Verification ------------------------------------------------------------
    @staticmethod
    def _check_shape_consistency(iterable, prefix: str) -> None:
        """Ensure that each array in `iterable` has the same shape."""
        shape = np.shape(iterable[0])
        if any(np.shape(A) != shape for A in iterable):
            raise ValueError(f"{prefix} shapes inconsistent")

    # Properties --------------------------------------------------------------
    @property
    def state_dimension(self) -> int:
        r"""Dimension :math:`r` of the state :math:`\qhat` that the operator
        acts on.
        """
        return None if self.entries is None else self.entries[0].shape[0]

    @property
    def parameter_dimension(self) -> int:
        r"""Dimension :math:`p` of the parameter vector :math:`\bfmu` that the
        operator matrix depends on.
        """
        return self.__p

    @parameter_dimension.setter
    def parameter_dimension(self, p):
        """Set :attr:`parameter_dimension`.
        Only allowed if :attr:`parameter_dimension` is currently ``None``.
        """
        if self.__p is not None:
            raise AttributeError(
                "can't set property 'parameter_dimension' twice"
            )
        if not isinstance(p, int) or p < 1:
            raise ValueError("parameter_dimension must be a positive integer")
        self.__p = p

    @property
    def shape(self) -> tuple:
        """Shape of the operator matrix when evaluated
        at a parameter value.
        """
        return None if self.entries is None else self.entries[0].shape

    @property
    def entries(self):
        r"""Arrays that define the operator matrix as a function of the
        parameter vector.
        """
        return self.__entries

    @abc.abstractmethod
    def set_entries(self, entries, fromblock: bool = False) -> None:
        r"""Set the arrays that define the operator matrix as a function of
        the parameter vector.

        Parameters
        ----------
        entries : list of s (r, d) ndarrays, or (r, sd) ndarray
            Operator entries, either as a list of arrays
            (``fromblock=False``, default)
            or as a horizontal concatenatation of arrays (``fromblock=True``).
        fromblock : bool
            If ``True``, interpret ``entries`` as a horizontal concatenation
            of arrays; if ``False`` (default), interpret ``entries`` as a list
            of arrays.
        """
        self.__entries = entries

    # Operator inference ------------------------------------------------------
    @abc.abstractmethod
    def operator_dimension(self, s: int, r: int, m: int = None) -> int:
        r"""Number of columns in the total operator matrix.

        Parameters
        ----------
        s : int
            Number of training parameter values.
        r : int
            State dimension.
        m : int or None
            Input dimension.

        Returns
        -------
        d : int
            Number of columns in the total operator entries matrix.
            This is also the number of rows in the data matrix block.
        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractmethod
    def datablock(self, parameters, states, inputs=None):
        r"""Return the data matrix block corresponding to the operator.

        Parameters
        ----------
        parameters : (s, p) ndarray
            Traning parameter values :math:`\bfmu_{0},\ldots,\bfmu_{s-1}`.
        states : list of s (r, k_i) ndarrays
            State snapshots for each of the :math:`s` training parameter
            values.
        inputs : list of s (m, k_i) ndarrays
            Inputs corresponding to the state snapshots.

        Returns
        -------
        block : (D, K) ndarray
            Data block for the parametric operator.
            Here, :math:`D` is the total operator matrix dimension and
            :math:`K = \sum_{i=0}^{s-1}k_i`, the total number of state
            snapshots.
        """
        raise NotImplementedError  # pragma: no cover


def is_opinf(obj) -> bool:
    """Return ``True`` if ``obj`` is an OpInf operator (whether or not the
    entries are calibrated).
    """
    return isinstance(obj, (OpInfOperator, ParametricOpInfOperator))


def is_uncalibrated(obj) -> bool:
    """Return ``True`` if ``obj`` is an OpInf operator with empty entries."""
    return is_opinf(obj) and obj.entries is None



# rom/_bayes.py
"""Classes supporting Bayesian operator inference."""

__all__ = [
    "OperatorPosterior",
    "BayesianROM",
]

import warnings
import numpy as np
import scipy.linalg
import scipy.stats

from .. import errors, lstsq, post, utils
from ..models import _utils as modutils
from ._base import _identity
from ._nonparametric import ROM


VALID_SOLVERS = (
    lstsq.L2Solver,
    lstsq.L2DecoupledSolver,
    lstsq.TikhonovSolver,
    lstsq.TikhonovDecoupledSolver,
)


# Posterior ===================================================================
class OperatorPosterior:
    r"""Posterior distribution for operator matrices.

    Operator inference models are uniquely determined by operator matrices
    :math:`\Ohat\in\RR^{r\times d}` that concatenate the entries of all
    operators in the model. For example, the time-continuous model

    .. math::
       \ddt\qhat(t) = \chat + \Ahat\qhat(t) + \Hhat[\qhat(t)\otimes\qhat(t)]

    is uniquely determined by the operator matrix

    .. math::
       \Ohat = [~\chat~~\Ahat~~\Hhat~] \in \RR^{r \times d}.

    Typical *deterministic* operator inference learns a single operator matrix
    :math:`\Ohat` from state measurements, while *probabilistic* or *Bayesian*
    operator inference constructs a distribution of operator matrices,
    :math:`p(\Ohat)`. This class implements an operator matrix distribution
    where the rows of :math:`\Ohat` are multivariate Normal (Gaussian) random
    variables, i.e.,

    .. math::
       p(\ohat_{i}) = \mathcal{N}(\ohat_i\mid\bfmu_i,\bfSigma_i),
       \\
       \bfmu_i \in \RR^{d},
       \quad
       \bfSigma_i \in \RR^{d\times d},
       \quad
       i = 0, \ldots, r-1,

    where :math:`\ohat_i \in \RR^{d}` is the :math:`i`-th row of :math:`\Ohat`.

    The :class:`BayesianROM` class has a ``posterior`` attribute that is an
    ``OperatorPosterior`` object.

    Parameters
    ----------
    means : list of r (d,) ndarrays
        Mean values for each row of the operator matrix.
    precisions : list of r (d, d) ndarrays
        **INVERSE** covariance matrices for each row of the operator matrix.
    alreadyinverted : bool
        If ``True``, assume ``precisions`` is the collection of covariance
        matrices, not their inverses.
    """

    def __init__(self, means, precisions, *, alreadyinverted=False):
        """Store and pre-process the distribution parameters."""
        if (r := len(means)) != (_r2 := len(precisions)):
            raise ValueError(f"len(means) = {r} != {_r2} = len(precisions)")

        self.__r = r
        self.__randomvariables = []

        for i in range(self.__r):
            # Verify dimensions.
            mean_i, cov_i = means[i], precisions[i]
            if not isinstance(mean_i, np.ndarray) or mean_i.ndim != 1:
                raise ValueError(f"means[{i}] should be a 1D ndarray")
            if not isinstance(cov_i, np.ndarray) or cov_i.ndim != 2:
                raise ValueError(f"precisions[{i}] should be a 2D ndarray")
            d = mean_i.shape[0]
            if cov_i.shape != (d, d):
                raise ValueError(f"means[{i}] and precisions[{i}] not aligned")

            # Make a multivariate Normal distribution for this operator row.
            if not alreadyinverted:
                cov_i = scipy.stats.Covariance.from_precision(cov_i)
            self.__randomvariables.append(
                scipy.stats.multivariate_normal(mean=mean_i, cov=cov_i)
            )

        # If operator rows are all the same size, wrap rvs() output as array.
        self.__rvsasarray = False
        d = means[0].size
        if all(mean.size == d for mean in means):
            self.__rvsasarray = True

    # Properties --------------------------------------------------------------
    @property
    def nrows(self) -> int:
        """Number of rows :math:`r` in the data matrix. This is also the state
        dimension of the corresponding model.
        """
        return self.__r

    @property
    def randomvariables(self) -> list:
        """Multivariate normal random variables for the rows of the operator
        matrix.
        """
        return self.__randomvariables

    @property
    def means(self) -> list:
        r"""Mean vectors :math:`\bfmu_0,\ldots,\bfmu_{r-1}\in\RR^{d}` for the
        rows of the operator matrix.
        """
        return [rv.mean for rv in self.randomvariables]

    @property
    def covs(self) -> list:
        r"""Covariance matrices
        :math:`\bfSigma_0,\ldots,\bfSigma_{r-1}\in\RR^{d\times d}`
        for the rows of the operator matrix.
        """
        return [rv.cov for rv in self.randomvariables]

    def __eq__(self, other):
        if self.nrows != other.nrows:
            return False
        for m1, m2 in zip(self.means, other.means):
            if m1.shape != m2.shape or not np.all(m1 == m2):
                return False
        for C1, C2 in zip(self.covs, other.covs):
            if C1.shape != C2.shape or not np.all(C1 == C2):
                return False
        return True

    # Random draws ------------------------------------------------------------
    def rvs(self):
        r"""Draw a random operator matrix from the posterior operator
        distribution.

        Returns
        -------
        Ohat : (r, d) ndarray
            Operator matrix sampled from :math:`p(\Ohat)`.
        """
        ohats = [rv.rvs()[0] for rv in self.randomvariables]
        return np.array(ohats) if self.__rvsasarray else ohats

    # Model persistance -------------------------------------------------------
    def save(self, savefile, overwrite=True):
        """Save the posterior operator distribution.

        Parameters
        ----------
        savefile : str
            File to save data to.
        overwrite : bool
            If False and ``savefile`` exists, raise an exception.
        """
        with utils.hdf5_savehandle(savefile, overwrite) as hf:
            hf.create_dataset("state_dimension", data=[self.nrows])
            for i, (mean_i, cov_i) in enumerate(zip(self.means, self.covs)):
                hf.create_dataset(f"means_{i}", data=mean_i)
                hf.create_dataset(f"covs_{i}", data=cov_i)

    @classmethod
    def load(cls, loadfile):
        """Load a previously saved posterior operator distribution.

        Parameters
        ----------
        loadfile : str
            File to load data from.
        """
        with utils.hdf5_loadhandle(loadfile) as hf:
            r = int(hf["state_dimension"][0])
            means = [hf[f"means_{i}"][:] for i in range(r)]
            covs = [hf[f"covs_{i}"][:] for i in range(r)]

        return cls(means, covs, alreadyinverted=True)


# Bayesian ROMs ===============================================================
class _BayesianROMMixin:
    """Mixin for ROM classes with a Bayesian operator posterior."""

    def __init__(self):
        self.__posterior = None
        self._validate_model_solver(self.model)

    @staticmethod
    def _validate_model_solver(model):
        if modutils.is_interpolatory(model):
            raise AttributeError(
                "Fully interpolatory parametric models are not supported "
                "for Bayesian ROMs"
            )
        if not hasattr(model, "solver") or not isinstance(
            model.solver, VALID_SOLVERS
        ):
            types = ", ".join(f"lstsq.{s.__name__}" for s in VALID_SOLVERS)
            raise AttributeError(
                "'model' must have a 'solver' attribute "
                f"of one of the following types: {types}"
            )

    @property
    def posterior(self) -> OperatorPosterior:
        """Posterior distribution for the operator matrices."""
        return self.__posterior

    def _initialize_posterior(self):
        """Set the operator posterior if numerically possible."""
        means, precisions = self.model.solver.posterior()
        try:
            self.__posterior = OperatorPosterior(means, precisions)
        except np.linalg.LinAlgError as ex:
            if ex.args[0] == "Matrix is not positive definite":
                self.__posterior = None

    def draw_operators(self):
        """Set the :attr:`model` operators to a new random draw from the
        :attr:`posterior` operator distribution.
        """
        self.model._extract_operators(self.posterior.rvs())

    def fit_regselect_continuous(
        self,
        candidates: list,
        train_time_domains: list,
        parameters: list,
        states: list,
        ddts: list = None,
        input_functions: list = None,
        fit_transformer: bool = True,
        fit_basis: bool = True,
        regularizer_factory=None,
        gridsearch_only: bool = False,
        test_time_length: float = 0,
        stability_margin: float = 5.0,
        num_posterior_draws: int = 20,
        test_cases: list = None,
        verbose: bool = False,
        **predict_options: dict,
    ):
        if not self._iscontinuous:
            raise AttributeError(
                "this method is for time-continuous models only, "
                "use fit_regselect_discrete()"
            )

        if parameters is None:
            states, ddts, input_functions = self._fix_single_trajectory(
                states, ddts, input_functions
            )

        # Validate arguments.
        if np.isscalar(train_time_domains[0]):
            train_time_domains = [train_time_domains] * len(states)
        for t, Q in zip(train_time_domains, states):
            if t.shape != (Q.shape[1],):
                raise errors.DimensionalityError(
                    "train_time_domains and states not aligned"
                )
        if input_functions is not None:
            if callable(input_functions):  # one global input function.
                input_functions = [input_functions] * len(states)
            if not callable(input_functions[0]):
                raise TypeError(
                    "argument 'input_functions' must be sequence of callables"
                )
            inputs = [  # evaluate the inputs over the time domain.
                np.column_stack([u(tt) for tt in t])
                for u, t in zip(input_functions, train_time_domains)
            ]
        else:
            inputs = None
        if test_time_length < 0:
            raise ValueError("argument 'test_time_length' must be nonnegative")
        if regularizer_factory is None:
            regularizer_factory = _identity
        processed_test_cases = self._process_test_cases(
            test_cases, utils.ContinuousRegTest
        )

        # Fit the model for the first time.
        states = self._fit_and_return_training_data(
            parameters=parameters,
            states=states,
            lhs=ddts,
            inputs=inputs,
            fit_transformer=fit_transformer,
            fit_basis=fit_basis,
        )

        # Set up the regularization selection.
        shifts, limits = self._get_stability_limits(states, stability_margin)

        def unstable(_Q, ell, size):
            """Return ``True`` if the solution is unstable."""
            if _Q.shape[-1] != size:
                return True
            return np.abs(_Q - shifts[ell]).max() > limits[ell]

        # Extend the training time domains by the testing time length.
        if test_time_length > 0:
            time_domains = []
            for t_train in train_time_domains:
                dt = np.mean(np.diff(t_train))
                t_test = t_train[-1] + np.linspace(
                    dt,
                    dt + test_time_length,
                    int(test_time_length / dt),
                )
                time_domains.append(np.concatenate(((t_train, t_test))))
        else:
            time_domains = train_time_domains

        if input_functions is None:
            input_functions = [None] * len(states)
        loop_collections = [states, input_functions, time_domains]
        if is_parametric := parameters is not None:
            loop_collections.insert(0, parameters)

        def update_model(reg_params):
            """Reset the regularizer and refit the model operators."""
            self.model.solver.regularizer = regularizer_factory(reg_params)
            self._initialize_posterior()

        def training_error(reg_params):
            """Compute the training error for a single regularization
            candidate by solving the model, checking for stability, and
            comparing to available training data.
            """
            update_model(reg_params)
            if self.posterior is None:
                return np.inf

            # Pass stability checks.
            for tcase in processed_test_cases:
                for _ in range(num_posterior_draws):
                    self.draw_operators()
                    if not tcase.evaluate(self.model, **predict_options):
                        return np.inf

            # Compute training error.
            error = 0
            for ell, entries in enumerate(zip(*loop_collections)):
                if is_parametric:
                    params, Q, input_func, t = entries
                    predict_args = (params, Q[:, 0], t, input_func)
                else:
                    Q, input_func, t = entries
                    predict_args = (Q[:, 0], t, input_func)
                draws = []
                trainsize = Q.shape[-1]
                for _ in range(num_posterior_draws):
                    self.draw_operators()
                    with warnings.catch_warnings():
                        warnings.simplefilter("ignore")
                        solution = self.model.predict(
                            *predict_args, **predict_options
                        )
                    if unstable(solution, ell, t.size):
                        return np.inf
                    draws.append(solution[:, :trainsize])
                solution_train = np.mean(draws, axis=0)
                error += post.Lp_error(Q, solution_train, t[:trainsize])[1]
            return error / len(states)

        best_regularization = utils.gridsearch(
            training_error,
            candidates,
            gridsearch_only=gridsearch_only,
            label="regularization",
            verbose=verbose,
        )

        update_model(best_regularization)
        return self

    def fit_regselect_discrete(
        self,
        candidates: list,
        parameters: list,
        states: list,
        inputs: list = None,
        fit_transformer: bool = True,
        fit_basis: bool = True,
        regularizer_factory=None,
        gridsearch_only: bool = False,
        num_test_iters: int = 0,
        stability_margin: float = 5.0,
        num_posterior_draws: int = 20,
        test_cases: list = None,
        verbose: bool = False,
    ):
        if self._iscontinuous:
            raise AttributeError(
                "this method is for fully discrete models only, "
                "use fit_regselect_continuous()"
            )

        if parameters is None:
            states, _, inputs = self._fix_single_trajectory(
                states, None, inputs
            )

        # Validate arguments.
        if num_test_iters < 0:
            raise ValueError(
                "argument 'num_test_iters' must be a nonnegative integer"
            )
        if inputs is not None:
            if len(inputs) != len(states):
                raise errors.DimensionalityError(
                    f"{len(states)} state trajectories but "
                    f"{len(inputs)} input trajectories detected"
                )
            for Q, U in zip(states, inputs):
                if U.shape[-1] < Q.shape[1] + num_test_iters:
                    raise ValueError(
                        "argument 'inputs' must contain enough data for "
                        f"{num_test_iters} iterations after the training data"
                    )
        if regularizer_factory is None:
            regularizer_factory = _identity
        processed_test_cases = self._process_test_cases(
            test_cases, utils.DiscreteRegTest
        )

        # Fit the model for the first time.
        states = self._fit_and_return_training_data(
            parameters=parameters,
            states=states,
            lhs=None,
            inputs=inputs,
            fit_transformer=fit_transformer,
            fit_basis=fit_basis,
        )

        # Set up the regularization selection.
        shifts, limits = self._get_stability_limits(states, stability_margin)

        def unstable(_Q, ell):
            """Return ``True`` if the solution is unstable."""
            if np.isnan(_Q).any() or np.isinf(_Q).any():
                return True
            return np.any(np.abs(_Q - shifts[ell]).max() > limits[ell])

        # Extend the iteration counts by the number of testing iterations.
        num_iters = [Q.shape[-1] for Q in states]
        if num_test_iters > 0:
            num_iters = [n + num_test_iters for n in num_iters]

        if inputs is None:
            inputs = [None] * len(states)
        loop_collections = [states, inputs, num_iters]
        if is_parametric := parameters is not None:
            loop_collections.insert(0, parameters)

        def update_model(reg_params):
            """Reset the regularizer and refit the model operators."""
            self.model.solver.regularizer = regularizer_factory(reg_params)
            self._initialize_posterior()

        def training_error(reg_params):
            """Compute the mean training error for a single regularization
            candidate by solving the model, checking for stability, and
            comparing to available training data.
            """
            update_model(reg_params)
            if self.posterior is None:
                return np.inf

            # Pass stability checks.
            for tcase in processed_test_cases:
                for _ in range(num_posterior_draws):
                    self.draw_operators()
                    if not tcase.evaluate(self.model):
                        return np.inf

            # Solve the model, check for stability, and compute training error.
            error = 0
            for ell, entries in enumerate(zip(*loop_collections)):
                if is_parametric:
                    params, Q, U, niter = entries
                    predict_args = (params, Q[:, 0], niter, U)
                else:
                    Q, U, niter = entries
                    predict_args = (Q[:, 0], niter, U)
                draws = []
                for _ in range(num_posterior_draws):
                    self.draw_operators()
                    with warnings.catch_warnings():
                        warnings.simplefilter("ignore")
                        solution = self.model.predict(*predict_args)
                    if unstable(solution, ell):
                        return np.inf
                    draws.append(solution[:, : Q.shape[-1]])
                error += post.frobenius_error(Q, np.mean(draws, axis=0))[1]
            return error / len(states)

        best_regularization = utils.gridsearch(
            training_error,
            candidates,
            gridsearch_only=gridsearch_only,
            label="regularization",
            verbose=verbose,
        )

        update_model(best_regularization)
        return self


class BayesianROM(ROM, _BayesianROMMixin):
    r"""Probabilistic nonparametric reduced-order model.

    This class connects classes from the various submodules to form a complete
    reduced-order modeling workflow for probabilistic models.

    High-dimensional data
    :math:`\to` transformed / preprocessed data
    :math:`\to` compressed data
    :math:`\to` low-dimensional probabilistic model.

    Operator inference models are uniquely determined by operator matrices
    :math:`\Ohat\in\RR^{r\times d}` that concatenate the entries of all
    operators in the model. For example, the time-continuous model

    .. math::
       \ddt\qhat(t) = \chat + \Ahat\qhat(t) + \Hhat[\qhat(t)\otimes\qhat(t)]

    is uniquely determined by the operator matrix

    .. math::
       \Ohat = [~\chat~~\Ahat~~\Hhat~] \in \RR^{r \times d}.

    Typical *deterministic* operator inference learns a single operator matrix
    :math:`\Ohat` from state measurements, while *probabilistic* or *Bayesian*
    operator inference constructs a distribution of operator matrices,
    :math:`p(\Ohat)`. This class solves a Bayesian linear inference to define
    an :class:`OperatorPosterior` and facilitates sampling from the posterior.
    See :cite:`guo2022bayesopinf`.

    Parameters
    ----------
    model : :mod:`opinf.models` object
        Nonparametric system model, an instance of one of the following:

        * :class:`opinf.models.ContinuousModel`
        * :class:`opinf.models.DiscreteModel`

        The model must have a ``solver`` of one of the following types:

        * :class:`opinf.lstsq.L2Solver`
        * :class:`opinf.lstsq.L2DecoupledSolver`
        * :class:`opinf.lstsq.TikhonovSolver`
        * :class:`opinf.lstsq.TikhonovDecoupledSolver`

    lifter : :mod:`opinf.lift` object or None
        Lifting transformation.
    transformer : :mod:`opinf.pre` object or None
        Preprocesser.
    basis : :mod:`opinf.basis` object or None
        Dimensionality reducer.
    ddt_estimator : :mod:`opinf.ddt` object or None
        Time derivative estimator.
        Ignored if ``model`` is not time continuous.

    Notes
    -----
    The ``operators`` attribute of the :attr:`model` represents a single draw
    from the operator distribution and is modified every time
    :meth:`draw_operators` or :meth:`predict` are called.
    """

    def __init__(
        self,
        model,
        *,
        lifter=None,
        transformer=None,
        basis=None,
        ddt_estimator=None,
    ):
        ROM.__init__(
            self,
            model,
            lifter=lifter,
            transformer=transformer,
            basis=basis,
            ddt_estimator=ddt_estimator,
        )
        _BayesianROMMixin.__init__(self)

    def fit(
        self,
        states,
        lhs=None,
        inputs=None,
        fit_transformer: bool = True,
        fit_basis: bool = True,
    ):
        ROM.fit(
            self,
            states=states,
            lhs=lhs,
            inputs=inputs,
            fit_transformer=fit_transformer,
            fit_basis=fit_basis,
        )
        self._initialize_posterior()

    def fit_regselect_continuous(
        self,
        candidates: list,
        train_time_domains: list,
        states: list,
        ddts: list = None,
        input_functions: list = None,
        fit_transformer: bool = True,
        fit_basis: bool = True,
        regularizer_factory=None,
        gridsearch_only: bool = False,
        test_time_length: float = 0,
        stability_margin: float = 5,
        num_posterior_draws: int = 20,
        test_cases: list = None,
        verbose: bool = False,
        **predict_options: dict,
    ):
        """Calibrate the time-continuous model to training data, selecting the
        regularization hyperparameter(s) that minimize the sample mean training
        error while maintaining stability over the testing regime.

        This method requires the :attr:`model` to be time-continuous; use
        :meth:`fit_regselect_discrete` for discrete models. The
        ``model.solver.regularizer`` is repeatedly adjusted, and the operator
        :attr:`posterior` is recalibrated, until a best regularization is
        selected. Training error is measured by comparing training data to the
        sample mean of ``num_posterior_draws`` model predictions. Stability is
        required for each of the individual model predictions.
        See :cite:`mcquarrie2021combustion,guo2022bayesopinf`.

        Parameters
        ----------
        candidates : list of regularization hyperparameters
            Regularization hyperparameters to check before carrying out a
            derivative-free optimization.
        train_time_domains : list of s (k_i,) ndarrays
            Time domain corresponding to the training states.
        states : list of s (n, k_i) ndarrays
            State snapshots in the original state space. Each array
            ``states[i]`` is data corresponding to a different trajectory;
            each column ``states[i][:, j]`` is one snapshot.
        ddts : list of s (n, k_i) ndarrays or None
            Snapshot time derivative data. Each array ``ddts[i]`` are the time
            derivatives of ``states[i]``; each column ``ddts[i][:, j]``
            corresponds to the snapshot ``states[i][:, j]``. If ``None``
            (default), these are estimated using :attr:`ddt_estimator`.
        input_functions : list of s callables or None
            Input functions mapping time to input vectors. Only required if the
            :attr:`model` takes external inputs. Each ``input_functions[i]``
            is the function corresponding to ``states[i]``, and
            ``input_functions[i](train_time_domains[i][j])`` is the input
            vector corresponding to the snapshot ``states[i][:, j]``.
        fit_transformer : bool
            If ``True`` (default), calibrate the preprocessing transformation
            using the ``states``.
            If ``False``, assume the transformer is already calibrated.
        fit_basis : bool
            If ``True`` (default), calibrate the high-to-low dimensional
            mapping using the ``states``.
            If ``False``, assume the basis is already calibrated.
        regularizer_factory : callable or None
            Function mapping regularization hyperparameters to the full
            regularizer. Specifically, ``regularizer_factory(candidates[i])``
            will be assigned to ``model.solver.regularizer`` for each ``i``.
            If ``None`` (default), set ``regularizer_factory()`` to the
            identity function.
        gridsearch_only : bool
            If ``True``, stop after checking all regularization ``candidates``
            and do not follow up with optimization.
        test_time_length : float or None
            Amount of time after the training regime in which to require model
            stability.
        stability_margin : float
            Factor by which the predicted reduced states may deviate from the
            range of the training reduced states without the trajectory being
            classified as unstable.
        num_posterior_draws : int
            Number of draws from the operator :attr:`posterior` for stability
            checks and for estimating the sample mean of model predictions.
        test_cases : list of ContinuousRegTest objects
            Additional test cases for which the model is required to be stable.
            See :class:`opinf.utils.ContinuousRegTest`.
        verbose : bool
            If ``True``, print information during the regularization selection.
        predict_options : dict or None
            Extra arguments for :meth:`opinf.models.ContinuousModel.predict`.

        Notes
        -----
        If there is only one trajectory of training data (s = 1), ``states``
        may be provided as an (n, k) ndarray. In this case, it is assumed that
        ``ddts`` (if provided) is an (n, k) ndarray.

        The ``train_time_domains`` may be a single one-dimensional array, in
        which case it is assumed that each trajectory ``states[i]`` corresponds
        to the same time domain. Similarly, if ``input_functions`` is a single
        callable, it is assumed to be the input function for each trajectory.
        """
        return _BayesianROMMixin.fit_regselect_continuous(
            self,
            candidates=candidates,
            train_time_domains=train_time_domains,
            parameters=None,
            states=states,
            ddts=ddts,
            input_functions=input_functions,
            fit_transformer=fit_transformer,
            fit_basis=fit_basis,
            regularizer_factory=regularizer_factory,
            gridsearch_only=gridsearch_only,
            test_time_length=test_time_length,
            stability_margin=stability_margin,
            num_posterior_draws=num_posterior_draws,
            test_cases=test_cases,
            verbose=verbose,
            **predict_options,
        )

    def fit_regselect_discrete(
        self,
        candidates: list,
        states: list,
        inputs: list = None,
        fit_transformer: bool = True,
        fit_basis: bool = True,
        regularizer_factory=None,
        gridsearch_only: bool = False,
        num_test_iters: int = 0,
        stability_margin: float = 5,
        num_posterior_draws: int = 20,
        test_cases: list = None,
        verbose: bool = False,
    ):
        """Calibrate the fully discrete model to training data, selecting the
        regularization hyperparameter(s) that minimize the sample mean training
        error while maintaining stability over the testing regime.

        This method requires the :attr:`model` to be fully discrete; use
        :meth:`fit_regselect_continuous` for time-continuous models. The
        ``model.solver.regularizer`` is repeatedly adjusted, and the operator
        :attr:`posterior` is recalibrated, until a best regularization is
        selected. Training error is measured by comparing training data to the
        sample mean of ``num_posterior_draws`` model predictions. Stability is
        required for each of the individual model predictions.
        See :cite:`mcquarrie2021combustion,guo2022bayesopinf`.

        Parameters
        ----------
        candidates : list of regularization hyperparameters
            Regularization hyperparameters to check. If a single hyperparameter
            is given, use it as the start of an optimization-based search.
        states : list of s (n, k_i) ndarrays
            State snapshots in the original state space. Each array
            ``states[i]`` is data corresponding to a different trajectory;
            each column ``states[i][:, j]`` is one snapshot. This method
            assumes the snapshots are sequential, i.e., the model maps
            ``states[i][:, j]`` to ``states[i][:, j+1]``.
        states : list of s (r, k_i) ndarrays
            State snapshots in the reduced state space. This method assumes
            the snapshots are sequential, i.e., the model maps
            ``states[i][:, j]`` to ``states[i][:, j+1]``.
        inputs : list of s (m, k_i + num_test_iters) ndarrays
            Inputs corresponding to the training data, together with inputs
            for the testing regime. Only required if the :attr:`model` takes
            external inputs.
        regularizer_factory : callable or None
            Function mapping regularization hyperparameters to the full
            regularizer. Specifically, ``regularizer_factory(candidates[i])``
            will be assigned to ``model.solver.regularizer`` for each ``i``.
        gridsearch_only : bool
            If ``True``, stop after checking all regularization ``candidates``
            and do not follow up with optimization.
        num_test_iters : int
            Number of iterations after the training data in which to require
            model stability.
        stability_margin : float,
            Factor by which the reduced states may deviate from the range of
            the training data without being flagged as unstable.
        num_posterior_draws : int
            Number of draws from the operator :attr:`posterior` for stability
            checks and for estimating the sample mean of model predictions.
        test_cases : list of DiscreteRegTest objects
            Additional test cases for which the model is required to be stable.
            See :class:`opinf.utils.DiscreteRegTest`.
        verbose : bool
            If ``True``, print information during the regularization selection.

        Notes
        -----
        If there is only one trajectory of training data (s = 1), ``states``
        may be provided as an (n, k) ndarray. In this case, it is assumed that
        ``inputs`` (if provided) is a single (m, k + num_test_iters) ndarray.
        """
        return _BayesianROMMixin.fit_regselect_discrete(
            self,
            candidates=candidates,
            parameters=None,
            states=states,
            inputs=inputs,
            fit_transformer=fit_transformer,
            fit_basis=fit_basis,
            regularizer_factory=regularizer_factory,
            gridsearch_only=gridsearch_only,
            num_test_iters=num_test_iters,
            stability_margin=stability_margin,
            num_posterior_draws=num_posterior_draws,
            test_cases=test_cases,
            verbose=verbose,
        )

    def predict(self, state0, *args, **kwargs):
        """Draw from the operator posterior and evaluate the resulting model.

        Arguments are the same as the ``predict()`` method of :attr:`model`.

        Parameters
        ----------
        state0 : (n,) ndarray
            Initial state, expressed in the original state space.
        *args : list
            Other positional arguments to the ``predict()`` method of
            :attr:`model`.
        **kwargs : dict
            Keyword arguments to the ``predict()`` method of :attr:`model`.

        Returns
        -------
        states: (n, k) ndarray
            Solution to the drawn model, expressed in the original state space.
        """
        self.draw_operators()
        return ROM.predict(self, state0, *args, **kwargs)



# lstsq/_tikhonov.py
"""Operator Inference least-squares solvers with Tikhonov regularization."""

__all__ = [
    "L2Solver",
    "L2DecoupledSolver",
    "TikhonovSolver",
    "TikhonovDecoupledSolver",
]

import abc
import types
import warnings
import numpy as np
import scipy.linalg as la
import scipy.sparse as sparse

from .. import errors, utils
from ..operators import _utils as oputils
from ._base import SolverTemplate, _require_trained


def _symmetrize(A):
    return (A.T + A) / 2


def _check_sigmas(sigmas):
    if np.any(sigmas < np.finfo(np.float64).eps):
        raise RuntimeError("zero residual --> posterior is deterministic")
    return sigmas


# Solver classes ==============================================================
class _BaseRegularizedSolver(SolverTemplate):
    r"""Base class for solvers of regularized linear least-squares problems.

    .. math::
       \argmin_{\ohat_i}
       \|\D\ohat_i - \z_i\|_2^2
       + \sum_{i=1}^{r}\|\bfGamma_i\ohat_i\|_2^2,
       \quad i = 1, \ldots, r.

    For each :math:`i`, this is equivalent to the following stacked ordinary
    least-squares problem:

    .. math::
       \argmin{\Ohat}
       \left\|
           \left[\begin{array}{c}\D \\ \bfGamma_i\end{array}\right]\ohat_i
           - \left[\begin{array}{c}\z_i \\ \0\end{array}\right]
        \right\|_2^2.

    The exact solution is described by the normal equations:

    .. math::
        (\D\trp\D + \bfGamma_i\trp\bfGamma_i)\ohat_i = \D\trp\z_i.
    """

    # Properties: regularization ----------------------------------------------
    @abc.abstractmethod
    def regularizer(self):
        """Regularization scalar, matrix, or list of these."""
        raise NotImplementedError  # pragma: no cover

    # Main methods ------------------------------------------------------------
    def fit(self, data_matrix: np.ndarray, lhs_matrix: np.ndarray):
        r"""Verify dimensions and save the data matrices.

        Parameters
        ----------
        data_matrix : (k, d) ndarray
            Data matrix :math:`\D`.
        lhs_matrix : (r, k) or (k,) ndarray
            "Left-hand side" data matrix :math:`\Z` (not its transpose!).
            If one-dimensional, assume :math:`r = 1`.
        """
        SolverTemplate.fit(self, data_matrix, lhs_matrix)
        if self.k < self.d:
            warnings.warn(
                "non-regularized least-squares system is underdetermined",
                errors.OpInfWarning,
            )
        return self

    @abc.abstractmethod
    def posterior(self):
        r"""Solve the Bayesian operator inference regression, constructing the
        means and inverse covariances of probability distributions for the
        rows of an operator matrix posterior.

        See :cite:`guo2022bayesopinf` for details.

        Returns
        -------
        means : list of r (d,) ndarrays
            Mean vectors.
        precisions : list of r (d, d) ndarrays
            Inverse covariance matrices.
        """
        raise NotImplementedError  # pragma: no cover

    # Post-processing ---------------------------------------------------------
    @abc.abstractmethod
    def regcond(self) -> float:
        r"""Compute the :math:`2`-norm condition number of the regularized
        data matrix :math:`[~\D\trp~~\bfGamma\trp~]\trp.`
        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractmethod
    def regresidual(self, Ohat: np.ndarray) -> np.ndarray:
        """Compute the residual of the regularized regression problem."""
        raise NotImplementedError  # pragma: no cover

    # Persistence -------------------------------------------------------------
    def _save(self, savefile, overwrite=False, extras=tuple()):
        """Serialize the solver, saving it in HDF5 format.
        The model can be recovered with the :meth:`_load()` class method.

        Parameters
        ----------
        savefile : str
            File to save to, with extension ``.h5`` (HDF5).
        overwrite : bool
            If ``True`` and the specified ``savefile`` already exists,
            overwrite the file. If ``False`` (default) and the specified
            ``savefile`` already exists, raise an error.
        extras : list
            Names of additional attributes to save.
        """
        with utils.hdf5_savehandle(savefile, overwrite) as hf:
            reg = self.regularizer
            if self.__class__ is L2Solver:
                reg = [reg]
            hf.create_dataset("regularizer", data=reg)

            if isinstance(self, TikhonovSolver):
                meta = hf.create_dataset("meta", shape=(0,))
                meta.attrs["method"] = self.method

            self._save_dict(hf, "options")

            if self.data_matrix is not None:
                hf.create_dataset("data_matrix", data=self.data_matrix)
                hf.create_dataset("lhs_matrix", data=self.lhs_matrix)
                for attr in extras:
                    hf.create_dataset(attr, data=getattr(self, attr))

    @classmethod
    def _load(cls, loadfile: str, extras=tuple()):
        """Load a serialized solver from an HDF5 file, created previously from
        the :meth:`save()` method.

        Parameters
        ----------
        loadfile : str
            Path to the file where the model was stored via :meth:`save()`.
        extras : list
            Names of additional attributes to load.

        Returns
        -------
        solver : _BaseRegularizedSolver
            Loaded solver.
        """
        with utils.hdf5_loadhandle(loadfile) as hf:

            reg = hf["regularizer"][:]
            if cls is L2Solver:
                reg = reg[0]

            options = cls._load_dict(hf, "options")
            kwargs = dict(
                regularizer=reg,
                lapack_driver=options["lapack_driver"],
            )

            if issubclass(cls, TikhonovSolver):
                if "cond" in options:
                    kwargs["cond"] = options["cond"]
                kwargs["method"] = hf["meta"].attrs["method"]

            solver = cls(**kwargs)

            if "data_matrix" in hf:
                D = hf["data_matrix"][:]
                Z = hf["lhs_matrix"][:]
                _BaseRegularizedSolver.fit(solver, D, Z)
                for attr in extras:
                    setattr(solver, attr, hf[attr][:])

        return solver


class L2Solver(_BaseRegularizedSolver):
    r"""Solve the Frobenius-norm ordinary least-squares problem with
    :math:`L_2` regularization.

    That is, solve

    .. math::
        \argmin_{\Ohat}\|\D\Ohat\trp - \Z\trp\|_F^2 + \|\lambda\Ohat\trp\|_F^2

    for some specified :math:`\lambda \ge 0`.

    The exact solution is described by the normal equations:

    .. math::
        (\D\trp\D + \lambda^2\I)\Ohat\trp = \D\trp\Z\trp,

    that is,

    .. math::
        \Ohat = \Z\D(\D\trp\D + \lambda^2\I)^{-\mathsf{T}}.

    Instead of solving these equations directly, the solution is calculated
    using the singular value decomposition of the data matrix :math:`\D`:
    if :math:`\D = \bfPhi\bfSigma\bfPsi\trp`, then
    :math:`\Ohat\trp = \bfPsi\bfSigma^{*}\bfPhi\trp\Z\trp` (i.e.,
    :math:`\Ohat = \Z\bfPhi\bfSigma^{*}\bfPsi\trp`), where
    :math:`\bfSigma^{*}` is a diagonal matrix with :math:`i`-th diagonal entry
    :math:`\Sigma_{i,i}^{*} = \Sigma_{i,i}/(\Sigma_{i,i}^{2} + \lambda^2).`

    Parameters
    ----------
    regularizer : float
        Scalar :math:`L_2` regularization constant :math:`\lambda`.
    lapack_driver : str
        LAPACK routine for computing the singular value decomposition.
        See :func:`scipy.linalg.svd()`.
    """

    def __init__(self, regularizer, lapack_driver: str = "gesdd"):
        """Store the regularizer and initialize attributes."""
        _BaseRegularizedSolver.__init__(self)
        self.regularizer = regularizer
        self.__options = types.MappingProxyType(
            dict(full_matrices=False, lapack_driver=lapack_driver)
        )

    # Properties --------------------------------------------------------------
    @property
    def regularizer(self):
        r"""Scalar :math:`L_2` regularization constant
        :math:`\lambda > 0.`
        """
        return self.__reg

    @regularizer.setter
    def regularizer(self, reg):
        """Set the regularization constant."""
        if not np.isscalar(reg):
            raise TypeError("regularization constant must be a scalar")
        if reg < 0:
            raise ValueError("regularization constant must be nonnegative")
        self.__reg = reg

    @property
    def options(self):
        """Keyword arguments for :func:`scipy.linalg.svd()`.
        These cannot be changed after instantiation.
        """
        return self.__options

    def __str__(self):
        """String representation: dimensions + solver options."""
        kwargs = self._print_kwargs(self.options)
        if np.isscalar(self.regularizer):
            regstr = f"{self.regularizer:.4e}"
        else:
            regstr = f"{self.regularizer.shape}"
        return "\n  ".join(
            [
                SolverTemplate.__str__(self),
                f"regularizer:     {regstr}",
                f"SVD solver:      scipy.linalg.svd({kwargs})",
            ]
        )

    # Main methods ------------------------------------------------------------
    def fit(self, data_matrix: np.ndarray, lhs_matrix: np.ndarray):
        r"""Verify dimensions and compute the singular value decomposition of
        the data matrix in preparation to solve the least-squares problem.

        Parameters
        ----------
        data_matrix : (k, d) ndarray
            Data matrix :math:`\D`.
        lhs_matrix : (r, k) or (k,) ndarray
            "Left-hand side" data matrix :math:`\Z` (not its transpose!).
            If one-dimensional, assume :math:`r = 1`.
        """
        _BaseRegularizedSolver.fit(self, data_matrix, lhs_matrix)

        Phi, svals, PsiT = la.svd(self.data_matrix, **self.options)
        self._svals = svals
        self._ZPhi = self.lhs_matrix @ Phi
        self._PsiT = PsiT

        return self

    @_require_trained
    def solve(self) -> np.ndarray:
        r"""Solve the Operator Inference regression.

        Returns
        -------
        Ohat : (r, d) ndarray
            Operator matrix :math:`\Ohat` (not its transpose!).
        """
        svals = self._svals.reshape((-1, 1))
        svals_inv = svals / (svals**2 + self.regularizer**2)
        return (self._ZPhi * svals_inv.T) @ self._PsiT

    def posterior(self):
        r"""Solve the Bayesian operator inference regression, constructing the
        means and inverse covariances of probability distributions for the
        rows of an operator matrix posterior.

        In this method, the :attr:`regularizer`, denoted :math:`\lambda`, is
        interpreted as a prior variance for the operator matrix distribution.
        The :math:`i`-th row of the operator matrix follows a multivariate
        normal distribution with the following mean and covariance.

        .. math::
           \bfmu_i &= \argmin_{\bfxi}\left\{
               \|\D\bfxi - \z_i\|_2^2 + \lambda^2\|\bfxi\|_2^2
           \right\},
           \\
           \bfSigma_i &= \sigma_i^2 \left(
               \D\trp\D + \lambda^2\I
           \right)^{-1},
           \\
           \sigma_i^2 &= \frac{1}{k}\left(
               \|\D\bfmu_i - \z_i\|_2^2 + \lambda^2\|\bfmu_i\|_2^2
           \right),

        where :math:`\z_i\in\RR^k` is the :math:`i`-th row of :math:`\Z`.
        See :cite:`guo2022bayesopinf` for details.

        Returns
        -------
        means : list of r (d,) ndarrays
            Mean vectors.
        precisions : list of r (d, d) ndarrays
            Inverse covariance matrices.

        Raises
        ------
        RuntimeError
            If any solver residual :math:`\sigma_i^2` is zero (a perfect
            regression), meaning the resulting covariance should be zero.
            In this case, do a deterministic regression with :meth:`solve()`,
            not a Bayesian regression.
        """
        Ohat = self.solve()
        DTD = _symmetrize(self.data_matrix.T @ self.data_matrix)
        invcov_unscaled = DTD + (self.regularizer**2 * np.eye(self.d))
        sigmas = _check_sigmas(self.regresidual(Ohat)) / self.k
        return Ohat, [invcov_unscaled / sig for sig in sigmas]

    # Post-processing ---------------------------------------------------------
    @_require_trained
    def cond(self):
        r"""Compute the :math:`2`-norm condition number of the data matrix
        :math:`\D`.
        """
        return self._svals.max() / self._svals.min()

    @_require_trained
    def regcond(self) -> float:
        r"""Compute the :math:`2`-norm condition number of the regularized
        data matrix :math:`[~\D\trp~~\lambda\I~]\trp.`

        Returns
        -------
        cond : float
            Condition number of the regularized data matrix.
        """
        svals2 = self._svals**2 + self.regularizer**2
        return np.sqrt(svals2.max() / svals2.min())

    @_require_trained
    def regresidual(self, Ohat: np.ndarray) -> np.ndarray:
        r"""Compute the residual of the regularized regression objective for
        each row of the given operator matrix.

        Specifically, given a potential :math:`\Ohat`, compute

        .. math::
           \|\D\ohat_i - \z_i\|_2^2 + \|\lambda\ohat_i\|_2^2,
           \quad i = 1, \ldots, r,

        where :math:`\ohat_i` and :math:`\z_i` are the :math:`i`-th rows of
        :math:`\Ohat` and :math:`\Z`, respectively.

        Parameters
        ----------
        Ohat : (r, d) ndarray
            Operator matrix :math:`\Ohat`.

        Returns
        -------
        residuals : (r,) ndarray
            :math:`2`-norm residuals for each row of the operator matrix.
        """
        residual = self.residual(Ohat)
        return residual + (self.regularizer**2 * np.sum(Ohat**2, axis=-1))

    # Persistence -------------------------------------------------------------
    def save(self, savefile: str, overwrite: bool = False):
        """Serialize the solver, saving it in HDF5 format.
        The model can be recovered with the :meth:`load()` class method.

        Parameters
        ----------
        savefile : str
            File to save to, with extension ``.h5`` (HDF5).
        overwrite : bool
            If ``True`` and the specified ``savefile`` already exists,
            overwrite the file. If ``False`` (default) and the specified
            ``savefile`` already exists, raise an error.
        """
        return self._save(savefile, overwrite, ["_svals", "_ZPhi", "_PsiT"])

    @classmethod
    def load(cls, loadfile: str):
        """Load a serialized solver from an HDF5 file, created previously from
        the :meth:`save()` method.

        Parameters
        ----------
        loadfile : str
            Path to the file where the model was stored via :meth:`save()`.

        Returns
        -------
        solver : L2Solver
            Loaded solver.
        """
        return cls._load(loadfile, ["_svals", "_ZPhi", "_PsiT"])

    def copy(self):
        """Make a copy of the solver."""
        solver = self.__class__(
            regularizer=self.regularizer,
            lapack_driver=self.options["lapack_driver"],
        )
        if self.data_matrix is not None:
            SolverTemplate.fit(solver, self.data_matrix, self.lhs_matrix)
            solver._svals = self._svals
            solver._ZPhi = self._ZPhi
            solver._PsiT = self._PsiT
        return solver


class L2DecoupledSolver(L2Solver):
    r"""Solve :math:`r` independent :math:`2`-norm ordinary least-squares
    problems, each with the same data matrix but a different :math:`L_2`
    regularization.

    That is, for :math:`i = 1, \ldots, r`, construct :math:`\Ohat` by solving

    .. math::
        \argmin_{\Ohat}\|\D\ohat_i - \z_i\|_2^2 + \|\lambda_i\ohat_i\|_2^2

    where :math:`\ohat_i` and :math:`\z_i` are the :math:`i`-th rows of
    :math:`\Ohat` and :math:`\Z`, respectively, with corresponding
    regularization constant :math:`\lambda_i > 0`.

    The exact solution for the :math:`i`-th problem is described by the normal
    equations:

    .. math::
        (\D\trp\D + \lambda_i^2\I)\ohat_i = \D\trp\z_i.

    Instead of solving these equations directly, the solution is calculated
    using the singular value decomposition of the data matrix
    (see :class:`L2Solver`).

    Parameters
    ----------
    regularizer : (r,) ndarray
        Scalar :math:`L_2` regularization constants, one for each row
        of the operator matrix.
    lapack_driver : str
        LAPACK routine for computing the singular value decomposition.
        See :func:`scipy.linalg.svd()`.
    """

    # Properties --------------------------------------------------------------
    def _check_regularizer_shape(self):
        if (shape1 := self.regularizer.shape) != (shape2 := (self.r,)):
            raise errors.DimensionalityError(
                f"regularizer.shape = {shape1} != {shape2} = (r,)"
            )

    @property
    def regularizer(self):
        r"""Scalar :math:`L_2` regularization constants, one for each row
        of the  operator matrix :math:`\Ohat`.
        """
        return self.__regs

    @regularizer.setter
    def regularizer(self, regs):
        """Set the regularization constants."""
        regs = np.array(regs)
        if regs.ndim != 1:
            raise ValueError("regularizer must be one-dimensional")
        if np.any(regs < 0):
            raise ValueError("regularization constants must be nonnegative")
        self.__regs = regs
        if self.r is not None:
            self._check_regularizer_shape()

    # Main methods ------------------------------------------------------------
    def fit(self, data_matrix: np.ndarray, lhs_matrix: np.ndarray):
        r"""Verify dimensions and compute the singular value decomposition of
        the data matrix in preparation to solve the least-squares problem.

        Parameters
        ----------
        data_matrix : (k, d) ndarray
            Data matrix :math:`\D`.
        lhs_matrix : (r, k) or (k,) ndarray
            "Left-hand side" data matrix :math:`\Z` (not its transpose!).
            If one-dimensional, assume :math:`r = 1`.
        """
        L2Solver.fit(self, data_matrix, lhs_matrix)
        self._check_regularizer_shape()
        return self

    def posterior(self):
        r"""Solve the Bayesian operator inference regression, constructing the
        means and inverse covariances of probability distributions for the
        rows of an operator matrix posterior.

        In this method, the :math:`i`-th entry of :attr:`regularizer`, denoted
        :math:`\lambda_i`, is interpreted as a prior variance for the
        :math:`i`-th row of the operator matrix, which follows a multivariate
        normal distribution with the following mean and covariance.

        .. math::
           \bfmu_i &= \argmin_{\bfxi}\left\{
               \|\D\bfxi - \z_i\|_2^2 + \lambda_i^2\|\bfxi\|_2^2
           \right\},
           \\
           \bfSigma_i &= \sigma_i^2 \left(
               \D\trp\D + \lambda_i^2\I
           \right)^{-1},
           \\
           \sigma_i^2 &= \frac{1}{k}\left(
               \|\D\bfmu_i - \z_i\|_2^2 + \lambda_i^2\|\bfmu_i\|_2^2
           \right),

        where :math:`\z_i\in\RR^k` is the :math:`i`-th row of :math:`\Z`.
        See :cite:`guo2022bayesopinf` for details.

        Returns
        -------
        means : list of r (d,) ndarrays
            Mean vectors.
        precisions : list of r (d, d) ndarrays
            Inverse covariance matrices.

        Raises
        ------
        RuntimeError
            If any solver residual :math:`\sigma_i^2` is zero (a perfect
            regression), meaning the resulting covariance should be zero.
            In this case, do a deterministic regression with :meth:`solve()`,
            not a Bayesian regression.
        """
        Ohat = self.solve()
        DTD = _symmetrize(self.data_matrix.T @ self.data_matrix)
        Id = np.eye(self.d)
        sigmas = _check_sigmas(self.regresidual(Ohat)) / self.k
        precisions = [
            (DTD + (reg**2 * Id)) / sig
            for reg, sig in zip(self.regularizer, sigmas)
        ]
        return Ohat, precisions

    # Post-processing ---------------------------------------------------------
    @_require_trained
    def regcond(self) -> float:
        r"""Compute the :math:`2`-norm condition number of each regularized
        data matrix, :math:`[~\D\trp~~\lambda_i\I~]\trp` for
        :math:`i = 1, \ldots, r`.

        Returns
        -------
        conds : (r,) ndarray
            Condition numbers of the regularized data matrices.
        """
        svals2 = self._svals**2 + self.regularizer.reshape((-1, 1)) ** 2
        return np.sqrt(svals2.max(axis=1) / svals2.min(axis=1))

    def regresidual(self, Ohat: np.ndarray) -> np.ndarray:
        r"""Compute the residual of the regularized regression objective for
        each row of the given operator matrix.

        Specifically, given a potential :math:`\Ohat`, compute

        .. math::
           \|\D\ohat_i - \z_i\|_2^2 + \|\lambda_i\ohat_i\|_2^2,
           \quad i = 1, \ldots, r,

        where :math:`\ohat_i` and :math:`\z_i` are the :math:`i`-th rows of
        :math:`\Ohat` and :math:`\Z`, respectively, and :math:`\lambda_i \ge 0`
        is the corresponding regularization constant.

        Parameters
        ----------
        Ohat : (r, d) ndarray
            Operator matrix :math:`\Ohat`.

        Returns
        -------
        residuals : (r,) ndarray
            :math:`2`-norm residuals for each row of the operator matrix.
        """
        return L2Solver.regresidual(self, Ohat)


class TikhonovSolver(_BaseRegularizedSolver):
    r"""Solve the Frobenius-norm ordinary least-squares problem with
    Tikhonov regularization.

    That is, solve

    .. math::
       \argmin_{\Ohat}\|\D\Ohat\trp - \Z\trp\|_F^2 + \|\bfGamma\Ohat\trp\|_F^2

    for a specified symmetric-positive-definite regularization matrix
    :math:`\bfGamma \in \RR^{d \times d}`. This is equivalent to solving the
    following stacked least-squares problem:

    .. math::
       \argmin_{\Ohat}
       \left\|
           \left[\begin{array}{c}\D \\ \bfGamma\end{array}\right]\Ohat\trp
           - \left[\begin{array}{c}\Z\trp \\ \0\end{array}\right]
       \right\|_F^2.

    The exact solution is described by the normal equations:

    .. math::
       (\D\trp\D + \bfGamma\trp\bfGamma)\Ohat\trp = \D\trp\Z\trp,

    that is,

    .. math::
       \Ohat = \Z\D(\D\trp\D + \bfGamma\trp\bfGamma)^{-\mathsf{T}}.

    Parameters
    ----------
    regularizer : (d, d) or (d,) ndarray
        Symmetric semi-positive-definite regularization matrix :math:`\bfGamma`
        or, if ``regularizer`` is a one-dimensional array, the diagonal entries
        of :math:`\bfGamma`. Here, ``d`` is the number of columns in the data
        matrix.
    method : str
        Strategy for solving the regularized least-squares problem.
        **Options**:

        * ``"lstsq"``: solve the stacked least-squares problem via
          :func:`scipy.linalg.lstsq()`; by default, this computes and uses the
          singular value decomposition of the stacked data matrix
          :math:`[~D\trp~~\bfGamma\trp~]\trp`.
        * ``"normal"``: directly solve the normal equations
          :math:`(\D\trp\D + \bfGamma\trp\bfGamma) \Ohat\trp = \D\trp\Z\trp`
          via :func:`scipy.linalg.solve()`.
    cond : float or None
        Cutoff for 'small' singular values of the data matrix,
        see :func:`scipy.linalg.lstsq()`. Ignored if ``method = "normal"``.
    lapack_driver : str or None
        Which LAPACK driver is used to solve the least-squares problem,
        see :func:`scipy.linalg.lstsq()`. Ignored if ``method = "normal"``.
    """

    def __init__(
        self,
        regularizer,
        method: str = "lstsq",
        cond: float = None,
        lapack_driver: str = None,
    ):
        """Store the regularizer and initialize attributes."""
        _BaseRegularizedSolver.__init__(self)
        self.regularizer = regularizer
        self.method = method
        self.__options = dict(cond=cond, lapack_driver=lapack_driver)

    # Properties --------------------------------------------------------------
    @property
    def options(self):
        """Keyword arguments for :func:`scipy.linalg.lstsq()`."""
        return self.__options

    def __str__(self):
        """String representation: dimensions + solver options."""
        kwargs = self._print_kwargs(self.options)
        if self.regularizer[0].ndim == 1:
            regstr = f"     {self.regularizer.shape}"
        else:
            regstr = (
                f"     {len(self.regularizer)} "
                f"{self.regularizer[0].shape} ndarrays"
            )
        if self.method == "lstsq":
            kwargs = self._print_kwargs(self.options)
            spstr = f"solver ('lstsq'): scipy.linalg.lstsq({kwargs})"
        else:
            spstr = "solver ('normal'): scipy.linalg.solve(assume_a='pos')"
        return "\n  ".join(
            [SolverTemplate.__str__(self), f"regularizer: {regstr}", spstr]
        )

    def _check_regularizer_shape(self):
        if (shape1 := self.regularizer.shape) != (shape2 := (self.d, self.d)):
            raise errors.DimensionalityError(
                f"regularizer.shape = {shape1} != {shape2} = (d, d)"
            )

    @property
    def regularizer(self):
        r"""Symmetric semi-positive-definite :math:`d \times d` regularization
        matrix :math:`\bfGamma`.
        """
        return self.__reg

    @regularizer.setter
    def regularizer(self, G):
        """Set the regularization matrix."""
        if sparse.issparse(G):
            G = G.toarray()
        elif not isinstance(G, np.ndarray):
            G = np.array(G)

        if G.ndim == 1:
            if np.any(G < 0):
                raise ValueError(
                    "diagonal regularizer must be positive semi-definite"
                )
            G = np.diag(G)

        self.__reg = G

        if self.d is not None:
            self._check_regularizer_shape()

    @classmethod
    def get_operator_regularizer(
        cls,
        operators: list,
        regularization_parameters: list,
        state_dimension: int,
        input_dimension: int = 0,
    ):
        r"""Construct a regularizer so that each operator is regularized
        separately.

        The regularization term for this solver is

        .. math::
           \|\bfGamma\Ohat\trp\|_F^2

        where :math:`\Ohat\in\RR^{r\times d}` is the unknown and
        :math:`\bfGamma\in\RR^{d \times d}` is a given regularization matrix.
        This method constructs :math:`\bfGamma` such that each operator
        represented in :math:`\Ohat` is regularized separately. For example, if
        :math:`\Ohat = [~\chat~~\Ahat~~\Hhat~~\Bhat~]`, then :math:`\bfGamma`
        may be designed so that

        .. math::
           \|\bfGamma\Ohat\trp\|_F^2
           = \gamma_1^2\|\chat\|_F^2
           + \gamma_2^2\|\Ahat\|_F^2
           + \gamma_3^2\|\Hhat\|_F^2
           + \gamma_4^2\|\Bhat\|_F^2.

        Parameters
        ----------
        operators : list of opinf.operators objects
            Collection of operators comprising the operator matrix.
        regularization_parameters : list of floats or ndarrays
            Regularization hyperparameters for each operator, i.e.,
            ``regularization_parameters[i]`` corresponds to ``operators[i]``.
        state_dimension : int
            Dimension of the (reduced) state.
        input_dimension : int
            Dimension of the input.
            If there is no input, this should be 0 (default).

        Returns
        -------
        regularizer : (d,) ndarray
            Diagonals of the regularization matrix so that ``operators[i]`` is
            regularized with constant ``regularization_parameters[i]``.

        Warns
        -----
        OpInfWarning
            If only one operator is provided (use :class:`L2Solver` instead),
            or if the ``input_dimension`` is provided but none of the operators
            act on inputs.

        Raises
        ------
        ValueError
            If a different number of ``operators`` and '
            ``regularization_parameters`` are provided, or if the
            ``input_dimension`` is not provided but at least one of the
            operators acts on inputs.
        TypeError
            If an entry ``operators`` has an unsupported type.
        """
        if (n1 := len(operators)) != (n2 := len(regularization_parameters)):
            raise ValueError(
                f"len(operators) == {n1} != "
                f"{n2} == len(regularization_parameters)"
            )
        if n1 == 1:
            warnings.warn(
                "consider using L2Solver for models with only one operator",
                errors.OpInfWarning,
            )

        # Check if input_dimension is needed or not.
        has_inputs = [oputils.has_inputs(op) for op in operators]
        inputs_required = any(has_inputs)
        if inputs_required and input_dimension == 0:
            idx = np.argmax(has_inputs)
            raise ValueError(
                "argument 'input_dimension' required, "
                f"operators[{idx}] acts on inputs"
            )
        elif not inputs_required and input_dimension > 0:
            warnings.warn(
                "argument 'input_dimension' ignored, "
                "no operators act on inputs",
                errors.OpInfWarning,
            )

        # Get operator dimensions.
        r, m = state_dimension, input_dimension
        dims = []
        for op in operators:
            if oputils.is_nonparametric(op):
                dims.append(op.operator_dimension(r, m))
            elif oputils.is_affine(op):
                dims.append(op.operator_dimension(None, r, m))
            else:
                raise TypeError(
                    f"unsupported operator type '{type(op).__name__}'"
                )

        # Construct the regularizer.
        regularizer = np.zeros(sum(dims))
        index = 0
        for dim, reg in zip(dims, regularization_parameters):
            endex = index + dim
            regularizer[index:endex] = reg
            index = endex

        return regularizer

    @property
    def method(self):
        """Strategy for solving the regularized least-squares problem, either
        ``"lstsq"`` (default) or ``"normal"``.
        """
        return self.__method

    @method.setter
    def method(self, method):
        """Set the method and precompute stuff as needed."""
        if method not in ("lstsq", "normal"):
            raise ValueError("method must be 'lstsq' or 'normal'")
        self.__method = method

    # Main routines -----------------------------------------------------------
    def fit(self, data_matrix: np.ndarray, lhs_matrix: np.ndarray):
        r"""Verify dimensions and precompute quantities in preparation to
        solve the least-squares problem.

        Parameters
        ----------
        data_matrix : (k, d) ndarray
            Data matrix :math:`\D`.
        lhs_matrix : (r, k) or (k,) ndarray
            "Left-hand side" data matrix :math:`\Z` (not its transpose!).
        """
        _BaseRegularizedSolver.fit(self, data_matrix, lhs_matrix)
        self._check_regularizer_shape()
        D, Z = self.data_matrix, self.lhs_matrix

        # Pad lhs matrix for "svd" solve.
        self._ZtPad = np.vstack((Z.T, np.zeros((self.d, self.r))))

        # Precompute normal equations terms for "normal" solve.
        self._DtD = D.T @ D
        self._DtZt = D.T @ Z.T

        return self

    @_require_trained
    def solve(self) -> np.ndarray:
        r"""Solve the Operator Inference regression.

        Returns
        -------
        Ohat : (r, d) ndarray
            Operator matrix :math:`\Ohat` (not its transpose!).
        """
        if self.method == "lstsq":
            DPad = np.vstack((self.data_matrix, self.regularizer))
            Ohat = la.lstsq(DPad, self._ZtPad, **self.options)[0].T
        elif self.method == "normal":
            regD = self._DtD + (self.regularizer.T @ self.regularizer)
            Ohat = la.solve(regD, self._DtZt, assume_a="pos").T
        return Ohat

    def posterior(self):
        r"""Solve the Bayesian operator inference regression, constructing the
        means and inverse covariances of probability distributions for the
        rows of an operator matrix posterior.

        In this method, the :attr:`regularizer`, denoted :math:`\bfGamma`, is
        interpreted as a prior covariance for the operator matrix distribution.
        The :math:`i`-th row of the operator matrix follows a multivariate
        normal distribution with the following mean and covariance.

        .. math::
           \bfmu_i &= \argmin_{\bfxi}\left\{
               \|\D\bfxi - \z_i\|_2^2 + \|\bfGamma\bfxi\|_2^2
           \right\},
           \\
           \bfSigma_i &= \sigma_i^2 \left(
               \D\trp\D + \bfGamma\trp\bfGamma
           \right)^{-1},
           \\
           \sigma_i^2 &= \frac{1}{k}\left(
               \|\D\bfmu_i - \z_i\|_2^2 + \|\bfGamma\bfmu_i\|_2^2
           \right),

        where :math:`\z_i\in\RR^k` is the :math:`i`-th row of :math:`\Z`.
        See :cite:`guo2022bayesopinf` for details.

        Returns
        -------
        means : list of r (d,) ndarrays
            Mean vectors.
        precisions : list of r (d, d) ndarrays
            Inverse covariance matrices.

        Raises
        ------
        RuntimeError
            If any solver residual :math:`\sigma_i^2` is zero (a perfect
            regression), meaning the resulting covariance should be zero.
            In this case, do a deterministic regression with :meth:`solve()`,
            not a Bayesian regression.
        """
        Ohat = self.solve()
        DTD = _symmetrize(self.data_matrix.T @ self.data_matrix)
        GTG = _symmetrize(self.regularizer.T @ self.regularizer)
        invcov_unscaled = DTD + GTG
        sigmas = _check_sigmas(self.regresidual(Ohat)) / self.k
        return Ohat, [invcov_unscaled / sig for sig in sigmas]

    # Post-processing ---------------------------------------------------------
    @_require_trained
    def regcond(self) -> float:
        r"""Compute the :math:`2`-norm condition number of the regularized
        data matrix :math:`[~\D\trp~~\bfGamma\trp~]\trp.`

        Returns
        -------
        cond : float
            Condition number of the regularized data matrix.
        """
        return np.linalg.cond(np.vstack((self.data_matrix, self.regularizer)))

    @_require_trained
    def regresidual(self, Ohat: np.ndarray) -> np.ndarray:
        r"""Compute the residual of the regularized regression objective for
        each row of the given operator matrix.

        Specifically, given a potential :math:`\Ohat`, compute

        .. math::
           \|\D\ohat_i - \z_i\|_2^2 + \|\bfGamma\ohat_i\|_2^2,
           \quad i = 1, \ldots, r,

        where :math:`\ohat_i` and :math:`\z_i` are the :math:`i`-th rows of
        :math:`\Ohat` and :math:`\Z`, respectively.

        Parameters
        ----------
        Ohat : (r, d) ndarray
            Operator matrix :math:`\Ohat`.

        Returns
        -------
        residuals : (r,) ndarray
            :math:`2`-norm residuals for each row of the operator matrix.
        """
        residual = self.residual(Ohat)
        return residual + np.sum((self.regularizer @ Ohat.T) ** 2, axis=0)

    def save(self, savefile: str, overwrite: bool = False):
        """Serialize the solver, saving it in HDF5 format.
        The model can be recovered with the :meth:`load()` class method.

        Parameters
        ----------
        savefile : str
            File to save to, with extension ``.h5`` (HDF5).
        overwrite : bool
            If ``True`` and the specified ``savefile`` already exists,
            overwrite the file. If ``False`` (default) and the specified
            ``savefile`` already exists, raise an error.
        """
        return self._save(savefile, overwrite, ["_ZtPad", "_DtD", "_DtZt"])

    @classmethod
    def load(cls, loadfile: str):
        """Load a serialized solver from an HDF5 file, created previously from
        the :meth:`save()` method.

        Parameters
        ----------
        loadfile : str
            Path to the file where the model was stored via :meth:`save()`.

        Returns
        -------
        solver : L2Solver
            Loaded solver.
        """
        return cls._load(loadfile, ["_ZtPad", "_DtD", "_DtZt"])

    def copy(self):
        """Make a copy of the solver."""
        solver = self.__class__(
            regularizer=self.regularizer,
            method=self.method,
            cond=self.options["cond"],
            lapack_driver=self.options["lapack_driver"],
        )
        if self.data_matrix is not None:
            SolverTemplate.fit(solver, self.data_matrix, self.lhs_matrix)
            solver._ZtPad = self._ZtPad
            solver._DtD = self._DtD
            solver._DtZt = self._DtZt
        return solver


class TikhonovDecoupledSolver(TikhonovSolver):
    r"""Solve :math:`r` independent :math:`2`-norm ordinary least-squares
    problems, each with the same data matrix but a different Tikhonov
    regularization.

    That is, for :math:`i = 1, \ldots, r`, construct :math:`\Ohat` by solving

    .. math::
        \argmin_\Ohat\|\D\Ohat\trp - \Z\trp\|_F^2 + \|\bfGamma\Ohat\trp\|_F^2

    where :math:`\ohat_i` and :math:`\z_i` are the :math:`i`-th rows of
    :math:`\Ohat` and :math:`\Z`, respectively, with corresponding
    symmetric-positive-definite regularization matrix :math:`\bfGamma_i`.

    This is equivalent to solving the following stacked least-squares problems:

    .. math::
       \argmin{\Ohat}
       \left\|
           \left[\begin{array}{c}\D \\ \bfGamma_i\end{array}\right]\Ohat\trp
           - \left[\begin{array}{c}\Z\trp \\ \0\end{array}\right]
       \right\|_F^2,
       \quad i = 1, \ldots, r.

    The exact solution of the :math:`i`-th problem is described by the normal
    equations:

    .. math::
       (\D\trp\D + \bfGamma_i\trp\bfGamma_i)\ohat_i = \D\trp\z_i.

    Parameters
    ----------
    regularizer : list of r (d, d) or (d,) ndarrays
        Symmetric semi-positive-definite regularization matrices
        :math:`\bfGamma_1,\ldots,\bfGamma_r`.
        If the ``i``th entry of ``regularizer`` is a one-dimensional array,
        it is intepreted as the diagonal entries of :math:`\bfGamma_i`. Here,
        `d` is the number of columns in the data matrix.
    method : str
        Strategy for solving the regularized least-squares problem.
        **Options**:

        * ``"lstsq"``: solve the stacked least-squares problem via
          :func:`scipy.linalg.lstsq()`; by default, this computes and uses the
          singular value decomposition of the stacked data matrix
          :math:`[~D\trp~~\bfGamma\trp~]\trp`.
        * ``"normal"``: directly solve the normal equations
          :math:`(\D\trp\D + \bfGamma\trp\bfGamma) \Ohat\trp = \D\trp\Z\trp`
          via :func:`scipy.linalg.solve()`.
    cond : float or None
        Cutoff for 'small' singular values of the data matrix,
        see :func:`scipy.linalg.lstsq()`. Ignored if ``method = "normal"``.
    lapack_driver : str or None
        Which LAPACK driver is used to solve the least-squares problem,
        see :func:`scipy.linalg.lstsq()`. Ignored if ``method = "normal"``.
    """

    # Properties --------------------------------------------------------------
    def _check_regularizer_shape(self):
        """Check that the regularizer has the correct shape."""
        if len(self.regularizer) != self.r:
            raise ValueError("len(regularizer) != r")
        for i, G in enumerate(self.regularizer):
            if (shape1 := G.shape) != (shape2 := (self.d, self.d)):
                raise ValueError(
                    f"regularizer[{i}].shape = {shape1} != {shape2} = (d, d)"
                )

    @property
    def regularizer(self):
        r"""Symmetric semi-positive-definite regularization matrices
        :math:`\bfGamma_1,\ldots,\bfGamma_r`, one for each row of the
        operator matrix.
        """
        return self.__regs

    @regularizer.setter
    def regularizer(self, Gs):
        """Set the regularization matrices."""
        regs = []
        for G in Gs:
            if sparse.issparse(G):
                G = G.toarray()
            elif not isinstance(G, np.ndarray):
                G = np.array(G)
            if G.ndim == 1:
                if np.any(G < 0):
                    raise ValueError(
                        "diagonal regularizer must be positive semi-definite"
                    )
                G = np.diag(G)
            regs.append(G)

        self.__regs = regs
        if self.d is not None:
            self._check_regularizer_shape()

    # Main methods ------------------------------------------------------------
    @_require_trained
    def solve(self) -> np.ndarray:
        r"""Solve the Operator Inference regression.

        Returns
        -------
        Ohat : (r, d) ndarray
            Operator matrix :math:`\Ohat` (not its transpose!).
        """
        Ohat = np.empty((self.r, self.d))

        # Solve each independent regression problem (sequentially for now).
        for i, Gamma in enumerate(self.regularizer):
            if self.method == "lstsq":
                Dpad = np.vstack((self.data_matrix, Gamma))
                Ohat[i] = la.lstsq(Dpad, self._ZtPad[:, i])[0]
            elif self.method == "normal":
                regD = self._DtD + Gamma.T @ Gamma
                Ohat[i] = la.solve(regD, self._DtZt[:, i], assume_a="pos")
        return Ohat

    def posterior(self):
        r"""Solve the Bayesian operator inference regression, constructing the
        means and inverse covariances of probability distributions for the
        rows of an operator matrix posterior.

        In this method, the :math:`i`-th entry of :attr:`regularizer`, denoted
        :math:`\bfGamma_i`, is interpreted as a prior covariance for the
        :math:`i`-th row of the operator matrix, which follows a multivariate
        normal distribution with the following mean and covariance.

        .. math::
           \bfmu_i &= \argmin_{\bfxi}\left\{
               \|\D\bfxi - \z_i\|_2^2 + \|\bfGamma_i\bfxi\|_2^2
           \right\},
           \\
           \bfSigma_i &= \sigma_i^2 \left(
               \D\trp\D + \bfGamma_i\trp\bfGamma_i
           \right)^{-1},
           \\
           \sigma_i^2 &= \frac{1}{k}\left(
               \|\D\bfmu_i - \z_i\|_2^2 + \|\bfGamma_i\bfmu_i\|_2^2
           \right),

        where :math:`\z_i\in\RR^k` is the :math:`i`-th row of :math:`\Z`.
        See :cite:`guo2022bayesopinf` for details.

        Returns
        -------
        means : list of r (d,) ndarrays
            Mean vectors.
        precisions : list of r (d, d) ndarrays
            Inverse covariance matrices.

        Raises
        ------
        RuntimeError
            If any solver residual :math:`\sigma_i^2` is zero (a perfect
            regression), meaning the resulting covariance should be zero.
            In this case, do a deterministic regression with :meth:`solve()`,
            not a Bayesian regression.
        """
        Ohat = self.solve()
        DTD = _symmetrize(self.data_matrix.T @ self.data_matrix)
        sigmas = _check_sigmas(self.regresidual(Ohat)) / self.k
        precisions = [
            (DTD + _symmetrize(Gamma.T @ Gamma)) / sig
            for Gamma, sig in zip(self.regularizer, sigmas)
        ]
        return Ohat, precisions

    # Post-processing ---------------------------------------------------------
    @_require_trained
    def regcond(self) -> float:
        r"""Compute the :math:`2`-norm condition number of each regularized
        data matrix :math:`[~\D\trp~~\bfGamma_i\trp~]\trp,~~i = 1, \ldots, r`.

        Returns
        -------
        conds : (r,) ndarray
            Condition numbers for the regularized data matrices.
        """
        return np.array(
            [
                np.linalg.cond(np.vstack((self.data_matrix, G)))
                for G in self.regularizer
            ]
        )

    @_require_trained
    def regresidual(self, Ohat: np.ndarray) -> np.ndarray:
        r"""Compute the residual of the regularized regression objective for
        each row of the given operator matrix.

        Specifically, given a potential :math:`\Ohat`, compute

        .. math::
           \|\D\ohat_i - \z_i\|_2^2 + \|\bfGamma_i\ohat_i\|_2^2,
           \quad i = 1, \ldots, r,

        where :math:`\ohat_i` and :math:`\z_i` are the :math:`i`-th rows of
        :math:`\Ohat` and :math:`\Z`, respectively, and :math:`\bfGamma_i` is
        the corresponding symmetric-positive-definite regularization matrix.

        Parameters
        ----------
        Ohat : (r, d) ndarray
            Operator matrix :math:`\Ohat`.

        Returns
        -------
        residuals : (r,) ndarray
            :math:`2`-norm residuals for each row of the operator matrix.
        """
        residual = self.residual(Ohat)
        rg = [np.sum((G @ oi) ** 2) for G, oi in zip(self.regularizer, Ohat)]
        return residual + np.array(rg)
